# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import warnings

from selene.support.shared import config


class OldConfig:
    @property
    def timeout(self):
        warnings.warn(
            'selene.config.timeout is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.timeout

    @timeout.setter
    def timeout(self, value):
        warnings.warn(
            'selene.config.timeout is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.timeout = value

    @property
    def poll_during_waits(self):
        warnings.warn(
            'selene.config.poll_during_waits is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.poll_during_waits

    @poll_during_waits.setter
    def poll_during_waits(self, value):
        warnings.warn(
            'selene.config.poll_during_waits is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.poll_during_waits = value

    @property
    def base_url(self):
        warnings.warn(
            'selene.config.base_url is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.base_url

    @base_url.setter
    def base_url(self, value):
        warnings.warn(
            'selene.config.base_url is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.base_url = value

    @property
    def app_host(self):
        warnings.warn(
            'selene.config.app_host is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.base_url

    @app_host.setter
    def app_host(self, value):
        warnings.warn(
            'selene.config.app_host is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.base_url = value

    @property
    def cash_elements(self):
        warnings.warn(
            'selene.config.cash_elements is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.cash_elements

    @cash_elements.setter
    def cash_elements(self, value):
        warnings.warn(
            'selene.config.cash_elements is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.cash_elements = value

    @property
    def browser_name(self):
        warnings.warn(
            'selene.config.browser_name is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.browser_name

    @browser_name.setter
    def browser_name(self, value):
        warnings.warn(
            'selene.config.browser_name is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.browser_name = value

    @property
    def start_maximized(self):
        warnings.warn(
            'selene.config.start_maximized is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.start_maximized

    @start_maximized.setter
    def start_maximized(self, value):
        warnings.warn(
            'selene.config.start_maximized is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.start_maximized = value

    @property
    def hold_browser_open(self):
        warnings.warn(
            'selene.config.hold_browser_open is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.hold_browser_open

    @hold_browser_open.setter
    def hold_browser_open(self, value):
        warnings.warn(
            'selene.config.hold_browser_open is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.hold_browser_open = value

    @property
    def counter(self):
        warnings.warn(
            'selene.config.counter is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.counter

    @counter.setter
    def counter(self, value):
        warnings.warn(
            'selene.config.counter is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.counter = value

    @property
    def reports_folder(self):
        warnings.warn(
            'selene.config.reports_folder is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.reports_folder

    @reports_folder.setter
    def reports_folder(self, value):
        warnings.warn(
            'selene.config.reports_folder is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.reports_folder = value

    @property
    def desired_capabilities(self):
        warnings.warn(
            'selene.config.desired_capabilities is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        return config.desired_capabilities

    @desired_capabilities.setter
    def desired_capabilities(self, value):
        warnings.warn(
            'selene.config.desired_capabilities is deprecated, '
            'use `from selene.support.shared import config` import',
            DeprecationWarning,
        )
        config.desired_capabilities = value



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import warnings
from typing import Any

from selene.core import match as _match

# --- be.* conditions --- #
from selene.core.condition import Condition
from selene.core.entity import Element, Collection, Browser

visible: Condition[Element] = _match.element_is_visible.not_
hidden: Condition[Element] = _match.element_is_hidden.not_

present: Condition[Element] = _match.element_is_present.not_
in_dom: Condition[Element] = _match.element_is_present.not_
# todo: do we need both present and in_dom?
# todo: consider deprecating existing
existing: Condition[Element] = _match.element_is_present.not_

absent: Condition[Element] = _match.element_is_absent.not_

enabled: Condition[Element] = _match.element_is_enabled.not_
disabled: Condition[Element] = _match.element_is_disabled.not_

blank: Condition[Element] = _match.element_is_blank.not_


# --- have.* conditions --- #


def exact_text(value) -> Condition[Element]:
    return _match.element_has_exact_text(value).not_


# todo: consider accepting int
def text(partial_value) -> Condition[Element]:
    return _match.element_has_text(partial_value).not_


def attribute(name: str, value: str = None):
    if value:
        warnings.warn(
            'passing second argument is deprecated; use have.attribute(foo).value(bar) instead',
            DeprecationWarning,
        )
        return _match.element_has_attribute(name).value(value).not_

    original = _match.element_has_attribute(name)
    negated = original.not_

    def value(self, expected: str, ignore_case=False) -> Condition[Element]:
        return original.value(expected, ignore_case).not_

    def value_containing(
        self, expected: str, ignore_case=False
    ) -> Condition[Element]:
        return original.value_containing(expected, ignore_case).not_

    def values(self, *expected: str) -> Condition[Collection]:
        return original.values(*expected).not_

    def values_containing(self, *expected: str) -> Condition[Collection]:
        return original.values_containing(*expected).not_

    negated.value = value
    negated.value_containing = value_containing
    negated.values = values
    negated.values_containing = values_containing

    return negated


def js_property(name: str, value: str = None):
    if value:
        warnings.warn(
            'passing second argument is deprecated; use have.js_property(foo).value(bar) instead',
            DeprecationWarning,
        )
        return _match.element_has_js_property(name).value(value).not_

    original = _match.element_has_js_property(name)
    negated = original.not_

    def value(self, expected: str) -> Condition[Element]:
        return original.value(expected).not_

    def value_containing(self, expected: str) -> Condition[Element]:
        return original.value_containing(expected).not_

    def values(self, *expected: str) -> Condition[Collection]:
        return original.values(*expected).not_

    def values_containing(self, *expected: str) -> Condition[Collection]:
        return original.values_containing(*expected).not_

    negated.value = value
    negated.value_containing = value_containing
    negated.values = values
    negated.values_containing = values_containing

    return negated


def css_property(name: str, value: str = None):
    if value:
        warnings.warn(
            'passing second argument is deprecated; use have.css_property(foo).value(bar) instead',
            DeprecationWarning,
        )
        return _match.element_has_css_property(name).value(value).not_

    original = _match.element_has_css_property(name)
    negated = original.not_

    def value(self, expected: str) -> Condition[Element]:
        return original.value(expected).not_

    def value_containing(self, expected: str) -> Condition[Element]:
        return original.value_containing(expected).not_

    def values(self, *expected: str) -> Condition[Collection]:
        return original.values(*expected).not_

    def values_containing(self, *expected: str) -> Condition[Collection]:
        return original.values_containing(*expected).not_

    negated.value = value
    negated.value_containing = value_containing
    negated.values = values
    negated.values_containing = values_containing

    return negated


def value(text) -> Condition[Element]:
    return _match.element_has_value(text).not_


def value_containing(partial_text) -> Condition[Element]:
    return _match.element_has_value_containing(partial_text).not_


def css_class(name) -> Condition[Element]:
    return _match.element_has_css_class(name).not_


def tag(name: str) -> Condition[Element]:
    return _match.element_has_tag(name).not_


def tag_containing(name: str) -> Condition[Element]:
    return _match.element_has_tag_containing(name).not_


# *** SeleneCollection conditions ***


def size(number: int) -> Condition[Collection]:
    return _match.collection_has_size(number).not_


def size_less_than(number: int) -> Condition[Collection]:
    return _match.collection_has_size_less_than(number).not_


def size_less_than_or_equal(number: int) -> Condition[Collection]:
    return _match.collection_has_size_less_than_or_equal(number).not_


def size_greater_than(number: int) -> Condition[Collection]:
    return _match.collection_has_size_greater_than(number).not_


def size_at_least(number: int) -> Condition[Collection]:
    warnings.warn(
        'might be deprecated; use have.size_greater_than_or_equal instead',
        PendingDeprecationWarning,
    )
    return _match.collection_has_size_greater_than_or_equal(number).not_


def size_greater_than_or_equal(number: int) -> Condition[Collection]:
    return _match.collection_has_size_greater_than_or_equal(number).not_


# todo: consider accepting ints
def texts(*partial_values: str) -> Condition[Collection]:
    return _match.collection_has_texts(*partial_values).not_


def exact_texts(*values: str) -> Condition[Collection]:
    return _match.collection_has_exact_texts(*values).not_


def url(exact_value: str) -> Condition[Browser]:
    return _match.browser_has_url(exact_value).not_


def url_containing(partial_value: str) -> Condition[Browser]:
    return _match.browser_has_url_containing(partial_value).not_


def title(exact_value: str) -> Condition[Browser]:
    return _match.browser_has_title(exact_value).not_


def title_containing(partial_value: str) -> Condition[Browser]:
    return _match.browser_has_title_containing(partial_value).not_


def tabs_number(value: int) -> Condition[Browser]:
    return _match.browser_has_tabs_number(value).not_


def tabs_number_less_than(value: int) -> Condition[Browser]:
    return _match.browser_has_tabs_number_less_than(value).not_


def tabs_number_less_than_or_equal(value: int) -> Condition[Browser]:
    return _match.browser_has_tabs_number_less_than_or_equal(value).not_


def tabs_number_greater_than(value: int) -> Condition[Browser]:
    return _match.browser_has_tabs_number_greater_than(value).not_


def tabs_number_greater_than_or_equal(value: int) -> Condition[Browser]:
    return _match.browser_has_tabs_number_greater_than_or_equal(value).not_


def js_returned_true(script_to_return_bool: str) -> Condition[Browser]:
    warnings.warn(
        'might be deprecated; use have.js_returned(True, ...) instead',
        PendingDeprecationWarning,
    )
    return _match.browser_has_js_returned(True, script_to_return_bool).not_


def js_returned(expected: Any, script: str, *args) -> Condition[Browser]:
    return _match.browser_has_js_returned(expected, script, *args).not_



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import sys
from typing import List, TypeVar, Generic

if sys.version_info >= (3, 10):
    from collections.abc import Callable
else:
    from typing import Callable

from selene.core.exceptions import ConditionNotMatchedError
from selene.core.wait import Predicate, Lambda


E = TypeVar('E')
R = TypeVar('R')


class Condition(Generic[E], Callable[[E], None]):
    @classmethod
    def by_and(cls, *conditions):
        def fn(entity):
            for condition in conditions:
                condition.call(entity)

        return cls(' and '.join(map(str, conditions)), fn)

    @classmethod
    def by_or(cls, *conditions):
        def fn(entity):
            errors: List[Exception] = []
            for condition in conditions:
                try:
                    condition.call(entity)
                    return
                except Exception as e:
                    errors.append(e)
            raise AssertionError('; '.join(map(str, errors)))

        return cls(' or '.join(map(str, conditions)), fn)

    @classmethod
    def as_not(
        cls, condition: Condition[E], description: str = None
    ) -> Condition[E]:
        # todo: how will it work composed conditions?
        condition_words = str(condition).split(' ')
        is_or_have = condition_words[0]
        name = ' '.join(condition_words[1:])
        no_or_not = 'not' if is_or_have == 'is' else 'no'
        new_description = description or f'{is_or_have} {no_or_not} {name}'

        def fn(entity):
            try:
                condition.call(entity)
            except Exception:
                return
            raise ConditionNotMatchedError()  # todo: try to handle printing actual values here too...

        return cls(new_description, fn)

    # function throwIfNot<E>(predicate: (entity: E) => Promise<boolean>): Lambda<E, void> {
    #     return async (entity: E) => {
    #         if (!await predicate(entity)) {
    #             throw new ConditionNotMatchedError();
    #         }
    #     };
    # }

    @classmethod
    def raise_if_not(
        cls, description: str, predicate: Predicate[E]
    ) -> Condition[E]:
        def fn(entity: E) -> None:
            if not predicate(entity):
                raise ConditionNotMatchedError()

        # TODO: consider the following style as one more option:
        #
        # @condition(description)  # or: @Condition.describe(description)
        # def fn(entity: E) -> None:
        #    if not predicate(entity):
        #        raise ConditionNotMatchedError()
        #
        # return fn

        return cls(description, fn)

    @classmethod
    def raise_if_not_actual(
        cls, description: str, query: Lambda[E, R], predicate: Predicate[R]
    ) -> Condition[E]:
        def fn(entity: E) -> None:
            query_to_str = str(query)
            result = (
                query.__name__
                if query_to_str.startswith('<function')
                else query_to_str
            )
            actual = query(entity)
            if not predicate(actual):
                raise AssertionError(f'actual {result}: {actual}')

        return cls(description, fn)

    def __init__(self, description: str, fn: Lambda[E, None]):
        self._description = description
        self._fn = fn

    def call(self, entity: E) -> None:
        self.__call__(entity)

    @property
    def predicate(self) -> Lambda[E, bool]:
        def fn(entity):
            try:
                self.call(entity)
                return True
            except Exception as e:
                return False

        return fn

    @property
    def not_(self) -> Condition[E]:  # todo: better name?
        return self.__class__.as_not(self)

    def __call__(self, entity: E) -> None:
        return self._fn(entity)

    def __str__(self):
        # todo: consider changing has to have on the fly for CollectionConditions
        # todo: or changing in collection locator rendering `all` to `collection`
        # todo: or changing in match.* names from collection_has_* to all_have_*
        return self._description

    def and_(self, condition: Condition[E]) -> Condition[E]:
        return Condition.by_and(self, condition)

    def or_(self, condition: Condition[E]) -> Condition[E]:
        return Condition.by_or(self, condition)


def not_(condition_to_be_inverted: Condition):  # todo: do we really need it?
    return condition_to_be_inverted.not_



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import warnings

from abc import abstractmethod, ABC
from typing import TypeVar, Union, List, Dict, Any, Callable, Tuple

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.switch_to import SwitchTo
from selenium.webdriver.remote.webelement import WebElement

from selene.common.fp import pipe
from selene.core.configuration import Config
from selene.core.wait import Wait, Command, Query
from selene.core.condition import Condition, not_
from selene.core.locator import Locator

from selene.common.helpers import to_by, flatten, is_absolute_url
from selene.core.exceptions import TimeoutException, _SeleneError

E = TypeVar('E', bound='Assertable')
R = TypeVar('R')


class Assertable(ABC):
    @abstractmethod
    # TODO: shouldn't we type self too?
    #       see #generic-methods-and-generic-self
    #       at https://mypy.readthedocs.io/en/stable/generics.html
    # def should(self: E, condition: Condition[E]) -> E:
    def should(self, condition: Condition[E]) -> E:
        pass


# todo: try as Matchable(ABC) and check if autocomplete will still work
class Matchable(Assertable):
    @abstractmethod
    def wait_until(self, condition: Condition[E]) -> bool:
        pass

    @abstractmethod
    def matching(self, condition: Condition[E]) -> bool:
        pass


class Configured(ABC):
    @property
    @abstractmethod
    def config(self) -> Config:
        pass


class WaitingEntity(Matchable, Configured):
    def __init__(self, config: Config):
        self._config = config

    @property
    def wait(self) -> Wait[E]:
        return self.config.wait(self)

    def perform(self, command: Command[E]) -> E:
        """Useful to call external commands.

        Commands might be predefined in Selene:
            element.perform(command.js.scroll_into_view)
        or some custom defined by selene user:
            element.perform(my_action.triple_click)

        You might think that it will be useful to use these methods also in Selene internally
        in order to define built in commands e.g. in Element class, like:

            def click(self):
                return self.perform(Command('click', lambda element: element().click()))

        instead of:

            def click(self):
                self.wait.for_(Command('click', lambda element: element().click()))
                return self

        But so far, we use the latter version - though, less concise, but more explicit,
        making it more obvious that waiting is built in;)

        """
        self.wait.for_(command)
        return self

    # todo: what about `entity[query.something]` syntax over or in addition to `entity.get(query.something)` ?
    def get(self, query: Query[E, R]) -> R:
        return self.wait.for_(query)

    # --- Configured --- #

    @property
    def config(self) -> Config:
        return self._config

    # --- Assertable --- #

    def should(self, condition: Condition[E]) -> E:
        self.wait.for_(condition)
        return self

    # --- Matchable --- #

    def wait_until(self, condition: Condition[E]) -> bool:
        return self.wait.until(condition)

    def matching(self, condition: Condition[E]) -> bool:
        return condition.predicate(self)


class Element(WaitingEntity):
    @staticmethod
    def _log_webelement_outer_html_for(
        element: Element,
    ) -> Callable[[TimeoutException], Exception]:
        def log_webelement_outer_html(error: TimeoutException) -> Exception:
            from selene.core import query
            from selene.core.match import element_is_present

            cached = element.cached

            if cached.matching(element_is_present):
                return TimeoutException(
                    error.msg
                    + f'\nActual webelement: {query.outer_html(element)}'
                )
            else:
                return error

        return log_webelement_outer_html

    # todo: should we move locator based init and with_ to Located base abstract class?

    def __init__(self, locator: Locator[WebElement], config: Config):
        self._locator = locator
        super().__init__(config)

    # --- Configured --- #

    def with_(self, config: Config = None, **config_as_kwargs) -> Element:
        return Element(
            self._locator, self.config.with_(config, **config_as_kwargs)
        )

    # --- Located --- #

    def __str__(self):
        return str(self._locator)

    def __call__(self) -> WebElement:
        return self._locator()

    # --- WaitingEntity --- #

    @property
    def wait(self) -> Wait[E]:
        # todo:  will not it break code like browser.with_(timeout=...)?
        # todo: fix that will disable/break shared hooks (snapshots)
        # return Wait(self,  # todo:  isn't it slower to create it each time from scratch? move to __init__?
        #             at_most=self.config.timeout,
        #             or_fail_with=pipe(
        #                 Element._log_webelement_outer_html_for(self),
        #                 self.config.hook_wait_failure))
        if self.config.log_outer_html_on_failure:
            """
            todo: remove this part completely from core.entity logic
                  move it to support.shared.config
            """
            return super().wait.or_fail_with(
                pipe(
                    Element._log_webelement_outer_html_for(self),
                    super().wait.hook_failure,
                )
            )
        else:
            return super().wait

    @property
    def cached(self) -> Element:
        # todo: do we need caching ? with lazy save of webelement to cache

        cache = None
        error = None
        try:
            cache = self()
        except Exception as e:
            error = e

        def get_webelement():
            if cache:
                return cache
            raise error

        return Element(
            Locator(f'{self}.cached', lambda: get_webelement()), self.config
        )

    # --- Relative location --- #

    def element(self, css_or_xpath_or_by: Union[str, tuple]) -> Element:
        by = to_by(css_or_xpath_or_by)

        return Element(
            Locator(f'{self}.element({by})', lambda: self().find_element(*by)),
            self.config,
        )

    def all(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        by = to_by(css_or_xpath_or_by)

        return Collection(
            Locator(f'{self}.all({by})', lambda: self().find_elements(*by)),
            self.config,
        )

    # --- Commands --- #

    def execute_script(self, script_on_self: str, *extra_args):
        driver: WebDriver = self.config.driver
        webelement = self()
        # todo: should we wrap it in wait or not?
        return driver.execute_script(script_on_self, webelement, *extra_args)

    # todo: do we need this method?
    #       do we really need to wrap script into function(element,args) here?
    #       if yes... wouldn't it be better to use standard arguments name
    #       instead of args?
    #       for better integration with js support in jetbrains products?
    def _execute_script(
        self,
        script_on_self_element_and_args: str,
        *extra_args,
    ):
        driver: WebDriver = self.config.driver
        webelement = self()
        # todo: should we wrap it in wait or not?
        return driver.execute_script(
            f'''
                return (function(element, args) {{
                    {script_on_self_element_and_args}
                }})(arguments[0], arguments[1])
            ''',
            webelement,
            extra_args,
        )

    def set_value(self, value: Union[str, int]) -> Element:
        # todo: should we move all commands like following or queries like in conditions - to separate py modules?
        # todo: should we make them webelement based (Callable[[WebElement], None]) instead of element based?
        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            webelement.clear()  # todo: change to impl based not on clear, because clear generates post-events...
            webelement.send_keys(str(value))

        from selene.core import command

        # todo: should we log the webelement source in the command name below?
        #       i.e. change from:
        #
        #   else Command(f'set value: {value}', fn)
        #
        #       to more low level:
        #
        #   else (
        #       Command(f'actual_not_overlapped_webelement.clear().send_keys({value})', fn)
        #       if self.config.wait_for_no_overlap_found_by_js
        #       else
        #       Command(f'actual_webelement.clear().send_keys({value})', fn)
        #   )
        #
        self.wait.for_(
            command.js.set_value(value)
            if self.config.set_value_by_js
            else Command(f'set value: {value}', fn)
        )

        # todo: consider returning self.cached, since after first successful call,
        #       all next ones should normally pass
        #       no waiting will be needed normally
        #       if yes - then we should pass fn commands to wait.for_ so the latter will return webelement to cache
        #       also it will make sense to make this behaviour configurable...
        return self

    def _actual_visible_webelement_and_maybe_its_cover(
        self, center_x_offset=0, center_y_offset=0
    ) -> Tuple[WebElement, WebElement]:
        # todo: will it be faster render outerHTML via lazy rendered SeleneError
        #       instead of: throw `element ${element.outerHTML} is not visible`
        #       in below js
        results = self._execute_script(
            '''
                var centerXOffset = args[0];
                var centerYOffset = args[1];

                var isVisible = !!(
                    element.offsetWidth
                    || element.offsetHeight
                    || element.getClientRects().length
                ) && window.getComputedStyle(element).visibility !== 'hidden'

                if (!isVisible) {
                    throw `element ${element.outerHTML} is not visible`
                }

                var rect = element.getBoundingClientRect();
                var x = rect.left + rect.width/2 + centerXOffset;
                var y = rect.top + rect.height/2 + centerYOffset;

                // TODO: now we return [element, null]
                //       in case of elementFromPoint returns null
                //       (kind of – if we don't know what to do,
                //       let's at least not block the execution...)
                //       rethink this... and handle the iframe case
                //       read more in
// https://developer.mozilla.org/en-US/docs/Web/API/Document/elementFromPoint

                var elementByXnY = document.elementFromPoint(x,y);
                if (elementByXnY == null) {
                    return [element, null];
                }

                var isNotOverlapped = element.isSameNode(elementByXnY);

                return isNotOverlapped
                       ? [element, null]
                       : [element, elementByXnY];
            ''',
            center_x_offset,
            center_y_offset,
        )
        webelement = results[0]
        maybe_cover = results[1]

        return webelement, maybe_cover

    @property
    def _actual_not_overlapped_webelement(self):
        (
            webelement,
            maybe_cover,
        ) = self._actual_visible_webelement_and_maybe_its_cover()
        if maybe_cover is not None:
            raise _SeleneError(
                lambda: f'Element: {webelement.get_attribute("outerHTML")}\n'
                + '\tis overlapped by: '
                + maybe_cover.get_attribute("outerHTML")
            )

        return webelement

    def type(self, text: Union[str, int]) -> Element:
        def fn(element: Element):
            if self.config.wait_for_no_overlap_found_by_js:
                webelement = element._actual_not_overlapped_webelement
            else:
                webelement = element()
            webelement.send_keys(str(text))

        from selene.core import command

        self.wait.for_(
            command.js.type(text)
            if self.config.type_by_js
            else Command(f'type: {text}', fn)
        )

        return self

    def press(self, *keys) -> Element:
        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            webelement.send_keys(*keys)

        self.wait.command(f'press keys: {keys}', fn)

        return self

    def press_enter(self) -> Element:
        return self.press(Keys.ENTER)

    def press_escape(self) -> Element:
        return self.press(Keys.ESCAPE)

    def press_tab(self) -> Element:
        return self.press(Keys.TAB)

    def clear(self) -> Element:
        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            webelement.clear()

        self.wait.command('clear', fn)

        return self

    def submit(self) -> Element:
        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            webelement.submit()

        self.wait.command('submit', fn)

        return self

    # todo: do we need config.click_by_js?
    # todo: add offset args with defaults, or add additional method, think on what is better
    def click(self) -> Element:
        """Just a normal click:)

        You might ask, why don't we have config.click_by_js?
        Because making all clicks js based is not a "normal case".
        You might need to speed up all "set value" in your tests, but command.js.click will not speed up anything.
        Yet, sometimes, it might be useful as workaround.
        In such cases - use

            element.perform(command.js.click)

        to achieve the goal in less concise way,
        thus, identifying by this "awkwardness" that it is really a workaround;)
        """
        self.wait.command('click', lambda element: element().click())
        return self

    def double_click(self) -> Element:
        actions: ActionChains = ActionChains(self.config.driver)

        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            actions.double_click(webelement).perform()

        self.wait.command('double click', fn)

        return self

    def context_click(self) -> Element:
        actions: ActionChains = ActionChains(self.config.driver)

        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            actions.context_click(webelement).perform()

        self.wait.command('context click', fn)

        return self

    def hover(self) -> Element:
        actions: ActionChains = ActionChains(self.config.driver)

        def fn(element: Element):
            webelement = (
                element._actual_not_overlapped_webelement
                if self.config.wait_for_no_overlap_found_by_js
                else element()
            )
            actions.move_to_element(webelement).perform()

        self.wait.command('hover', fn)

        return self

    # todo: should we reflect queries as self methods? or not...
    # pros: faster to query element attributes
    # cons: queries are not test oriented. test is steps + asserts
    #       so queries will be used only occasionally, then why to make a heap from Element?
    #       hence, occasionally it's enough to have them called as
    #           query.outer_html(element)  # non-waiting version
    #       or
    #           element.get(query.outer_html)  # waiting version
    # def outer_html(self) -> str:
    #     return self.wait.for_(query.outer_html)

    # --- Assertable --- #

    # we need this method here in order to make autocompletion work...
    # unfortunately the "base class" version is not enough
    def should(
        self, condition: Condition[Element], timeout: int = None
    ) -> Element:
        if timeout:
            warnings.warn(
                "using timeout argument is deprecated; "
                "use `browser.element('#foo').with_(Config(timeout=6)).should(be.enabled)`"
                "or just `...with_(timeout=6).should(...` style instead",
                DeprecationWarning,
            )
            return self.with_(Config(timeout=timeout)).should(condition)

        super().should(condition)
        return self

    # --- Deprecated --- #

    def get_actual_webelement(self) -> WebElement:
        warnings.warn(
            "considering to be deprecated; use element as callable instead, like: browser.element('#foo')()",
            PendingDeprecationWarning,
        )
        return self()

    def caching(self) -> Element:
        warnings.warn(
            "deprecated; use `cached` property instead: browser.element('#foo').cached",
            DeprecationWarning,
        )
        return self.cached

    def s(self, css_or_xpath_or_by: Union[str, tuple]) -> Element:
        # warnings.warn(
        #     "consider using more explicit `element` instead: browser.element('#foo').element('.bar')",
        #     SyntaxWarning)
        return self.element(css_or_xpath_or_by)

    def find(self, css_or_xpath_or_by: Union[str, tuple]) -> Element:
        warnings.warn(
            "deprecated; consider using `element` instead: browser.element('#foo').element('.bar')",
            DeprecationWarning,
        )
        return self.element(css_or_xpath_or_by)

    def ss(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        # warnings.warn(
        #     "consider using `all` instead: browser.element('#foo').all('.bar')",
        #     SyntaxWarning)
        return self.all(css_or_xpath_or_by)

    def find_all(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        warnings.warn(
            "deprecated; consider using `all` instead: browser.element('#foo').all('.bar')",
            DeprecationWarning,
        )
        return self.all(css_or_xpath_or_by)

    def elements(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        warnings.warn(
            "deprecated; consider using `all` instead: browser.element('#foo').all('.bar')",
            DeprecationWarning,
        )
        return self.all(css_or_xpath_or_by)

    @property
    def parent_element(self):
        warnings.warn(
            "considering to deprecate because such impl makes less sens for mobile automation; "
            "consider using corresponding xpath instead if you really need it for web:"
            "browser.element('#foo').element('..')",
            PendingDeprecationWarning,
        )
        return self.element('..')

    @property
    def following_sibling(self):
        warnings.warn(
            "considering to deprecate because such impl makes less sens for mobile automation; "
            "consider using corresponding xpath instead if you really need it for web:"
            "browser.element('#foo').element('./following-sibling::*')",
            PendingDeprecationWarning,
        )
        return self.element('./following-sibling::*')

    @property
    def first_child(self):
        warnings.warn(
            "considering to deprecate because such impl makes less sens for mobile automation; "
            "consider using corresponding xpath instead if you really need it for web: "
            "browser.element('#foo').element('./*[1]')",
            PendingDeprecationWarning,
        )
        return self.element('./*[1]')

    def assure(self, condition: ElementCondition, timeout=None) -> Element:
        warnings.warn(
            "deprecated; use `should` method instead: browser.element('#foo').should(be.enabled)",
            DeprecationWarning,
        )
        return self.should(condition, timeout)

    def should_be(self, condition: ElementCondition, timeout=None) -> Element:
        warnings.warn(
            "deprecated; use `should` method with `be.*` style conditions instead: "
            "browser.element('#foo').should(be.enabled)",
            DeprecationWarning,
        )
        return self.should(condition, timeout)

    def should_have(
        self, condition: ElementCondition, timeout=None
    ) -> Element:
        warnings.warn(
            "deprecated; use `should` method with `have.*` style conditions instead: "
            "browser.element('#foo').should(have.text('bar')",
            DeprecationWarning,
        )
        return self.should(condition, timeout)

    def should_not(self, condition: ElementCondition, timeout=None) -> Element:
        warnings.warn(
            "deprecated; use `should` method with `be.not_.*` or `have.no.*` style conditions instead: "
            "`browser.element('#foo').should(be.not_.enabled).should(have.no.css_class('bar')`, "
            "you also can build your own inverted conditions by: `not_ok = Condition.as_not(have.text('Ok'))`",
            DeprecationWarning,
        )
        return self.should(not_(condition), timeout)

    def assure_not(self, condition: ElementCondition, timeout=None) -> Element:
        warnings.warn(
            "deprecated; use `should` method with `be.not_.*` or `have.no.*` style conditions instead: "
            "`browser.element('#foo').should(be.not_.enabled).should(have.no.css_class('bar')`, "
            "you also can build your own inverted conditions by: `not_ok = Condition.as_not(have.text('Ok'))`",
            DeprecationWarning,
        )
        return self.should(not_(condition), timeout)

    def should_not_be(
        self, condition: ElementCondition, timeout=None
    ) -> Element:
        warnings.warn(
            "deprecated; use `should` method with `be.not_.*` or `have.no.*` style conditions instead: "
            "`browser.element('#foo').should(be.not_.enabled).should(have.no.css_class('bar')`, "
            "you also can build your own inverted conditions by: `not_ok = Condition.as_not(have.text('Ok'))`",
            DeprecationWarning,
        )
        return self.should(not_(condition), timeout)

    def should_not_have(
        self, condition: ElementCondition, timeout=None
    ) -> Element:
        warnings.warn(
            "deprecated; use `should` method with `be.not_.*` or `have.no.*` style conditions instead: "
            "`browser.element('#foo').should(be.not_.enabled).should(have.no.css_class('bar')`, "
            "you also can build your own inverted conditions by: `not_ok = Condition.as_not(have.text('Ok'))`",
            DeprecationWarning,
        )
        return self.should(not_(condition), timeout)

    def set(self, value: Union[str, int]) -> Element:
        warnings.warn(
            "deprecated; use `set_value` method instead", DeprecationWarning
        )
        return self.set_value(value)

    def scroll_to(self):
        warnings.warn(
            "deprecated; use `browser.element('#foo').perform(command.js.scroll_into_view)` style instead",
            DeprecationWarning,
        )
        from selene.core import command

        self.perform(command.js.scroll_into_view)
        return self

    def press_down(self):
        warnings.warn(
            "deprecated; use `browser.element('#foo').type(Keys.ARROW_DOWN)` style instead",
            DeprecationWarning,
        )
        return self.type(Keys.ARROW_DOWN)

    def find_element(self, by=By.ID, value=None):
        warnings.warn(
            "deprecated; use `browser.element('#foo').should(be.in_dom)().find_element(by, value)` style instead",
            DeprecationWarning,
        )
        self.wait.command(
            'find element', lambda element: element().find_element(by, value)
        )
        return self

    def find_elements(self, by=By.ID, value=None):
        warnings.warn(
            "deprecated; use `browser.element('#foo').should(be.in_dom)().find_elements(by, value)` style instead",
            DeprecationWarning,
        )
        self.wait.command(
            'find elements', lambda element: element().find_elements(by, value)
        )
        return self

    def send_keys(self, *value) -> Element:
        # warnings.warn(
        #     "send_keys is deprecated because the name is not user-oriented in context of web ui e2e tests; "
        #     "use `type` for 'typing text', press(*key) or press_* for 'pressing keys' methods instead",
        #     DeprecationWarning,
        # )
        # """
        # was deprecated previously
        # but deprecation was removed as an exception case
        # because sometimes send_keys is used for low level stuff
        # like sending values for files, etc...
        # let's rethink this topic later...
        # """
        self.wait.command(
            'send keys', lambda element: element().send_keys(*value)
        )
        return self

    @property
    def tag_name(self) -> str:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.tag_name)` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.tag)

    @property
    def text(self) -> str:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.text)` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.text)

    def attribute(self, name: str) -> str:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.attribute('name'))` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.attribute(name))

    def js_property(self, name: str) -> str:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.js_property('name'))` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.js_property(name))

    def value_of_css_property(self, name: str) -> str:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.css_property('name'))` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.css_property(name))

    def get_attribute(self, name: str) -> str:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.attribute('name'))` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.attribute(name))

    def get_property(self, name: str) -> str:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.js_property('name'))` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.js_property(name))

    def is_selected(self) -> bool:
        warnings.warn(
            "deprecated; use `browser.element('#foo').matching(be.selected)` style instead",
            DeprecationWarning,
        )
        from selene.support.conditions import be

        return self.matching(be.selected)

    def is_enabled(self):
        warnings.warn(
            "deprecated; use `browser.element('#foo').matching(be.enabled)` style instead",
            DeprecationWarning,
        )
        from selene.support.conditions import be

        return self.matching(be.enabled)

    def is_displayed(self):
        warnings.warn(
            "deprecated; use `browser.element('#foo').matching(be.visible)` style instead",
            DeprecationWarning,
        )
        from selene.support.conditions import be

        return self.matching(be.visible)

    @property
    def location(self) -> Dict[str, int]:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.location)` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.location)

    @property
    def location_once_scrolled_into_view(self) -> Dict[str, int]:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.location_once_scrolled_into_view)` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.location_once_scrolled_into_view)

    @property
    def size(self) -> Dict[str, Any]:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.size)` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.size)

    @property
    def rect(self) -> Dict[str, Any]:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.rect)` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.rect)

    @property
    def screenshot_as_base64(self) -> Any:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.screenshot_as_base64)` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.screenshot_as_base64)

    @property
    def screenshot_as_png(self) -> Any:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.screenshot_as_base64)` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.screenshot_as_png)

    def screenshot(self, filename: str) -> bool:
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.screenshot('filename'))` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.screenshot(filename))

    @property
    def parent(self):
        warnings.warn(
            "deprecated; use `browser.element('#foo')().parent` style for the majority of cases",
            DeprecationWarning,
        )
        return self.get(
            Query('parent search context', lambda element: element().parent)
        )

    @property
    def id(self):
        warnings.warn(
            "deprecated; use `browser.element('#foo').get(query.internal_id)` style instead",
            DeprecationWarning,
        )
        from selene.core import query

        return self.get(query.internal_id)


class SeleneElement(Element):  # todo: consider deprecating this name
    pass


class Collection(WaitingEntity):
    def __init__(self, locator: Locator[List[WebElement]], config: Config):
        self._locator = locator
        super().__init__(config)

    def with_(self, config: Config = None, **config_as_kwargs) -> Collection:
        return Collection(
            self._locator, self.config.with_(config, **config_as_kwargs)
        )

    def __str__(self):
        return str(self._locator)

    def __call__(self) -> List[WebElement]:
        return self._locator()

    @property
    def cached(self) -> Collection:
        webelements = self()
        return Collection(
            Locator(f'{self}.cached', lambda: webelements), self.config
        )

    def __iter__(self):
        i = 0
        cached = self.cached
        while i < len(cached()):
            element = cached[i]
            yield element
            i += 1

    def __len__(self):
        from selene.core import query

        return self.get(query.size)

    # todo: add config.index_collection_from_1, disabled by default
    def element(self, index: int) -> Element:
        def find() -> WebElement:
            webelements = self()
            length = len(webelements)

            if length <= index:
                raise AssertionError(
                    f'Cannot get element with index {index} '
                    + f'from webelements collection with length {length}'
                )

            return webelements[index]

        return Element(Locator(f'{self}[{index}]', find), self.config)

    @property
    def first(self):
        """
        in the past you used .first() was a method, now it's a property: .first
        probably when migrating to 2.* your code was broken here with an attribute error.
        just change .first() to .first everywhere in you code base;)
        """
        return self.element(0)

    def sliced(self, start: int, stop: int, step: int = 1) -> Collection:
        def find() -> List[WebElement]:
            webelements = self()

            # todo: assert length according to provided start, stop...

            return webelements[start:stop:step]

        return Collection(
            Locator(f'{self}[{start}:{stop}:{step}]', find), self.config
        )

    def __getitem__(
        self, index_or_slice: Union[int, slice]
    ) -> Union[Element, Collection]:
        if isinstance(index_or_slice, slice):
            return self.sliced(
                index_or_slice.start, index_or_slice.stop, index_or_slice.step
            )

        return self.element(index_or_slice)

    def from_(self, start: int) -> Collection:
        return self[start:]

    def to(self, stop: int) -> Collection:
        return self[:stop]

    def filtered_by(
        self, condition: Union[Condition[Element], Callable[[E], None]]
    ) -> Collection:
        condition = (
            condition
            if isinstance(condition, Condition)
            else Condition(str(condition), condition)
        )

        return Collection(
            Locator(
                f'{self}.filtered_by({condition})',
                lambda: [
                    element()
                    for element in self.cached
                    if element.matching(condition)
                ],
            ),
            self.config,
        )

    def filtered_by_their(
        self,
        selector_or_callable: Union[str, tuple, Callable[[Element], Element]],
        condition: Condition[Element],
    ) -> Collection:
        """
        :param selector_or_callable:
            - selector may be a str with css/xpath selector or tuple with by.* locator
            - callable should be a function on element that returns element
        :param condition: a condition to
        :return: collection subset with inner/relative element matching condition

        GIVEN html elements somewhere in DOM::
            .result
                .result-title
                .result-url
                .result-snippet

        THEN::

            browser.all('.result')\
                .filtered_by_their('.result-title', have.text('Selene'))\
                .should(have.size(3))

        ... is a shortcut for::

            browser.all('.result')\
                .filtered_by_their(lambda it: have.text(text)(it.element('.result-title')))\
                .should(have.size(3))

        OR with PageObject:

        THEN::

            results.element_by_its(lambda it: Result(it).title, have.text(text))\
                .should(have.size(3))

        Shortcut for::

            results.element_by(lambda it: have.text(text)(Result(it).title))\
                .should(have.size(3))

        WHERE::

            results = browser.all('.result')
            class Result:
                def __init__(self, element):
                    self.element = element
                    self.title = self.element.element('.result-title')
                    self.url = self.element.element('.result-url')
            # ...
        """
        warnings.warn(
            'filtered_by_their is experimental; might be renamed or removed in future',
            FutureWarning,
        )

        def find_in(parent: Element):
            if callable(selector_or_callable):
                return selector_or_callable(parent)
            else:
                return parent.element(selector_or_callable)

        return self.filtered_by(lambda it: condition(find_in(it)))

    def element_by(
        self, condition: Union[Condition[Element], Callable[[E], None]]
    ) -> Element:
        # todo: In the implementation below...
        #       We use condition in context of "matching", i.e. as a predicate...
        #       why then not accept Callable[[E], bool] also?
        #       (as you remember, Condition is Callable[[E], None] throwing Error)
        #       This will allow the following code be possible
        #           results.element_by(lambda it:
        #               Result(it).title.matching(have.text(text)))
        #       instead of:
        #           results.element_by(lambda it: have.text(text)(
        #                              Result(it).title))
        #       in addition to:
        #           results.element_by_its(lambda it:
        #               Result(it).title, have.text(text))
        #       Open Points:
        #       - do we need element_by_its, if we allow Callable[[E], bool] ?
        #       - if we add elements_by_its, do we need then to accept Callable[[E], bool] ?
        #       - probably... Callable[[E], bool] will lead to worse error messages,
        #         in such case we ignore thrown error's message
        #         - hm... ut seems like we nevertheless ignore it...
        #           we use element.matching(condition) below
        condition = (
            condition
            if isinstance(condition, Condition)
            else Condition(str(condition), condition)
        )

        def find() -> WebElement:
            cached = self.cached

            for element in cached:
                if element.matching(condition):
                    return element()

            from selene.core import query

            if self.config.log_outer_html_on_failure:
                """
                todo: move it support.shared.config
                """
                outer_htmls = [query.outer_html(element) for element in cached]

                raise AssertionError(
                    f'\n\tCannot find element by condition «{condition}» '
                    f'\n\tAmong {self}'
                    f'\n\tActual webelements collection:'
                    f'\n\t{outer_htmls}'
                )  # todo: isn't it better to print it all the time via hook, like for Element?
            else:
                raise AssertionError(
                    f'\n\tCannot find element by condition «{condition}» '
                    f'\n\tAmong {self}'
                )

        return Element(
            Locator(f'{self}.element_by({condition})', find), self.config
        )

    def element_by_its(
        self,
        selector_or_callable: Union[str, tuple, Callable[[Element], Element]],
        condition: Condition[Element],
    ) -> Element:
        """
        :param selector_or_callable:
            - selector may be a str with css/xpath selector or tuple with by.* locator
            - callable should be a function on element that returns element
        :param condition: a condition to
        :return: element from collection that has inner/relative element matching condition

        GIVEN html elements somewhere in DOM::
            .result
                .result-title
                .result-url
                .result-snippet

        THEN::

            browser.all('.result')\
                .element_by_its('.result-title', have.text(text))\
                .element('.result-url').click()

        ... is a shortcut for::

            browser.all('.result')\
                .element_by(lambda it: have.text(text)(it.element('.result-title')))\
                .element('.result-url').click()

        OR with PageObject:

        THEN::

            Result(results.element_by_its(lambda it: Result(it).title, have.text(text)))\
                .url.click()

        Shortcut for::

            Result(results.element_by(lambda it: have.text(text)(Result(it).title)))\
                .url.click()

        WHERE::

            results = browser.all('.result')
            class Result:
                def __init__(self, element):
                    self.element = element
                    self.title = self.element.element('.result-title')
                    self.url = self.element.element('.result-url')
            # ...
        """
        # todo: main questions to answer before removing warning:
        #       - isn't it enough to allow Callable[[Element], bool] as condition?
        #           browser.all('.result').element_by(
        #               lambda it: it.element('.result-title').matching(have.text('browser tests in Python')))
        #               .element('.result-url').click()
        #       - how to improve error messages in case we pass lambda (not a fun with good name/str repr)?
        #       - what about accepting collection condition? should we allow it?
        warnings.warn(
            'element_by_its is experimental; might be renamed or removed in future',
            FutureWarning,
        )

        def find_in(parent: Element):
            if callable(selector_or_callable):
                return selector_or_callable(parent)
            else:
                return parent.element(selector_or_callable)

        return self.element_by(lambda it: condition(find_in(it)))

    # todo: consider adding ss alias
    def all(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        warnings.warn(
            'might be renamed or deprecated in future; '
            'all is actually a shortcut for collected(lambda element: element.all(selector)...'
            'but we also have all_first and...'
            'it is yet unclear what name would be best for all_first as addition to all... '
            'all_first might confuse with all(...).first... I mean: '
            'all_first(selector) is actually '
            'collected(lambda e: e.element(selector)) '
            'but it is not the same as '
            'all(selector).first '
            'that is collected(lambda e: e.all(selector)).first ... o_O ',
            FutureWarning,
        )
        by = to_by(css_or_xpath_or_by)

        # todo: consider implement it through calling self.collected
        #       because actually the impl is self.collected(lambda element: element.all(selector))

        return Collection(
            Locator(
                f'{self}.all({by})',
                lambda: flatten(
                    [webelement.find_elements(*by) for webelement in self()]
                ),
            ),
            self.config,
        )

    # todo: consider adding s alias
    def all_first(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        warnings.warn(
            'might be renamed or deprecated in future; '
            'it is yet unclear what name would be best... '
            'all_first might confuse with all(...).first... I mean: '
            'all_first(selector) is actually '
            'collected(lambda e: e.element(selector)) '
            'but it is not the same as '
            'all(selector).first '
            'that is collected(lambda e: e.all(selector)).first ... o_O ',
            FutureWarning,
        )
        by = to_by(css_or_xpath_or_by)

        # todo: consider implement it through calling self.collected
        #       because actually the impl is self.collected(lambda element: element.element(selector))

        return Collection(
            Locator(
                f'{self}.all_first({by})',
                lambda: [
                    webelement.find_element(*by) for webelement in self()
                ],
            ),
            self.config,
        )

    def collected(
        self, finder: Callable[[Element], Union[Element, Collection]]
    ) -> Collection:
        # todo: consider adding predefined queries to be able to write
        #         collected(query.element(selector))
        #       over
        #         collected(lambda element: element.element(selector))
        #       and
        #         collected(query.all(selector))
        #       over
        #         collected(lambda element: element.all(selector))
        #       consider also putting such element builders like to find.* module instead of query.* module
        #       because they are not supposed to be used in entity.get(*) context defined for other query.* fns

        return Collection(
            Locator(
                f'{self}.collected({finder})',
                # todo: consider skipping None while flattening
                lambda: flatten(
                    [finder(element)() for element in self.cached]
                ),
            ),
            self.config,
        )

    # --- Assertable --- #

    def should(
        self,
        condition: Union[Condition[Collection], Condition[Element]],
        timeout: int = None,
    ) -> Collection:
        if isinstance(condition, ElementCondition):
            # todo: consider deprecating... makes everything too complicated...
            for element in self:
                if timeout:
                    warnings.warn(
                        "using timeout argument is deprecated; "
                        "use `browser.all('.foo').with_(Config(timeout=6)).should(have.size(0))`"
                        "or just `...with_(timeout=6).should(...` style instead",
                        DeprecationWarning,
                    )
                    element.with_(Config(timeout=timeout)).should(condition)
                element.should(condition)
        else:
            if timeout:
                warnings.warn(
                    "using timeout argument is deprecated; "
                    "use `browser.all('.foo').with_(Config(timeout=6)).should(have.size(0))` "
                    "or just `...with_(timeout=6).should(...` style instead",
                    DeprecationWarning,
                )
                self.with_(Config(timeout=timeout)).should(condition)
            super().should(condition)
        return self

    # --- Deprecated --- #

    def get_actual_webelements(self) -> List[WebElement]:
        warnings.warn(
            "considering to be deprecated; use collection as callable instead, like: browser.all('.foo')()",
            PendingDeprecationWarning,
        )
        return self()

    def caching(self) -> Collection:
        warnings.warn(
            "deprecated; use `cached` property instead: browser.all('#foo').cached",
            DeprecationWarning,
        )
        return self.cached

    def all_by(self, condition: Condition[Element]) -> Collection:
        warnings.warn(
            "deprecated; use `filtered_by` instead: browser.all('.foo').filtered_by(be.enabled)",
            DeprecationWarning,
        )
        return self.filtered_by(condition)

    def filter_by(self, condition: Condition[Element]) -> Collection:
        warnings.warn(
            "deprecated; use `filtered_by` instead: browser.all('.foo').filtered_by(be.enabled)",
            DeprecationWarning,
        )
        return self.filtered_by(condition)

    def find_by(self, condition: Condition[Element]) -> Element:
        warnings.warn(
            "deprecated; use `element_by` instead: browser.all('.foo').element_by(be.enabled)",
            DeprecationWarning,
        )
        return self.element_by(condition)

    def size(self):
        warnings.warn(
            "deprecated; use `len` standard function instead: len(browser.all('.foo'))",
            DeprecationWarning,
        )
        return len(self)

    def should_each(
        self, condition: ElementCondition, timeout=None
    ) -> Collection:
        # warnings.warn(
        #     "deprecated; use `should` method instead: browser.all('.foo').should(have.css_class('bar'))",
        #     DeprecationWarning,
        # )
        # """
        # was deprecated
        # but... probably making .should(condition) too smart was a bad idea...
        # TODO: decide on the the .should_each fate...
        # """
        return self.should(condition, timeout)

    def assure(
        self,
        condition: Union[CollectionCondition, ElementCondition],
        timeout=None,
    ) -> Collection:
        warnings.warn(
            "deprecated; use `should` method instead: browser.all('.foo').should(have.size(0))",
            DeprecationWarning,
        )
        return self.should(condition, timeout)

    def should_be(
        self,
        condition: Union[CollectionCondition, ElementCondition],
        timeout=None,
    ) -> Collection:
        warnings.warn(
            "deprecated; use `should` method with `be.*` style conditions instead: "
            "browser.all('.foo').should(be.*)",
            DeprecationWarning,
        )
        return self.should(condition, timeout)

    def should_have(
        self,
        condition: Union[CollectionCondition, ElementCondition],
        timeout=None,
    ) -> Collection:
        warnings.warn(
            "deprecated; use `should` method with `have.*` style conditions instead: "
            "browser.all('.foo').should(have.size(0))",
            DeprecationWarning,
        )
        return self.should(condition, timeout)

    def should_not(
        self,
        condition: Union[CollectionCondition, ElementCondition],
        timeout=None,
    ) -> Collection:
        warnings.warn(
            "deprecated; use `should` method with `be.not_.*` or `have.no.*` style conditions instead: "
            "`browser.all('.foo').should(have.no.size(2))`, "
            "you also can build your own inverted conditions by: `not_zero = Condition.as_not(have.size(0'))`",
            DeprecationWarning,
        )
        return self.should(not_(condition), timeout)

    def assure_not(
        self,
        condition: Union[CollectionCondition, ElementCondition],
        timeout=None,
    ) -> Collection:
        warnings.warn(
            "deprecated; use `should` method with `be.not_.*` or `have.no.*` style conditions instead: "
            "`browser.all('.foo').should(have.no.size(2))`, "
            "you also can build your own inverted conditions by: `not_zero = Condition.as_not(have.size(0'))`",
            DeprecationWarning,
        )
        return self.should(not_(condition), timeout)

    def should_not_be(
        self,
        condition: Union[CollectionCondition, ElementCondition],
        timeout=None,
    ) -> Collection:
        warnings.warn(
            "deprecated; use `should` method with `be.not_.*` or `have.no.*` style conditions instead: "
            "`browser.all('.foo').should(have.no.size(2))`, "
            "you also can build your own inverted conditions by: `not_zero = Condition.as_not(have.size(0'))`",
            DeprecationWarning,
        )
        return self.should(not_(condition), timeout)

    def should_not_have(
        self,
        condition: Union[CollectionCondition, ElementCondition],
        timeout=None,
    ) -> Collection:
        warnings.warn(
            "deprecated; use `should` method with `be.not_.*` or `have.no.*` style conditions instead: "
            "`browser.element('#foo').should(have.no.size(2))`, "
            "you also can build your own inverted conditions by: `not_zero = Condition.as_not(have.size(0'))`",
            DeprecationWarning,
        )
        return self.should(not_(condition), timeout)


class SeleneCollection(Collection):  # todo: consider deprecating this name
    pass


class Browser(WaitingEntity):
    def __init__(
        self, config: Config
    ):  # todo: what about adding **config_as_kwargs?
        super().__init__(config)

    # todo: consider implement it as context manager too...
    def with_(self, config: Config = None, **config_as_kwargs) -> Browser:
        return Browser(self.config.with_(config, **config_as_kwargs))

    def __str__(self):
        return 'browser'

    # todo: consider making it callable ...

    @property
    def driver(self) -> WebDriver:
        return self.config.driver

    # @property
    # def actions(self) -> ActionChains:
    #     """
    #     It's kind of just a shortcut for pretty low level actions from selenium webdriver
    #     Yet unsure about this property here:)
    #     comparing to usual high level Selene API...
    #     Maybe later it would be better to make own Actions with selene-style retries, etc.
    #     """
    #     return ActionChains(self.config.driver)

    # --- Element builders --- #

    def element(self, css_or_xpath_or_by: Union[str, tuple]) -> Element:
        by = to_by(css_or_xpath_or_by)

        return Element(
            Locator(
                f'{self}.element({by})', lambda: self.driver.find_element(*by)
            ),
            self.config,
        )

    def all(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        by = to_by(css_or_xpath_or_by)

        return Collection(
            Locator(
                f'{self}.all({by})', lambda: self.driver.find_elements(*by)
            ),
            self.config,
        )

    # --- High Level Commands--- #

    def open(self, relative_or_absolute_url: str) -> Browser:
        width = self.config.window_width
        height = self.config.window_height

        if width and height:
            self.driver.set_window_size(int(width), int(height))

        is_absolute = is_absolute_url(relative_or_absolute_url)
        base_url = self.config.base_url
        url = (
            relative_or_absolute_url
            if is_absolute
            else base_url + relative_or_absolute_url
        )

        self.driver.get(url)
        return self

    def switch_to_next_tab(self) -> Browser:
        from selene.core import query

        self.driver.switch_to.window(query.next_tab(self))

        # todo: should we user waiting version here (and in other similar cases)?
        # self.perform(Command(
        #     'open next tab',
        #     lambda browser: browser.driver.switch_to.window(query.next_tab(self))))

        return self

    def switch_to_previous_tab(self) -> Browser:
        from selene.core import query

        self.driver.switch_to.window(query.previous_tab(self))
        return self

    def switch_to_tab(self, index_or_name: Union[int, str]) -> Browser:
        if isinstance(index_or_name, int):
            index = index_or_name
            from selene.core import query

            self.driver.switch_to.window(query.tab(index)(self))
        else:
            self.driver.switch_to.window(index_or_name)

        return self

    @property
    def switch_to(self) -> SwitchTo:
        return self.driver.switch_to

    # todo: should we add also a shortcut for self.driver.switch_to.alert ?
    #       if we don't need to switch_to.'back' after switch to alert - then for sure we should...
    #       question is - should we implement our own alert as waiting entity?

    def close_current_tab(self) -> Browser:
        self.driver.close()
        return self

    def quit(self) -> None:
        self.driver.quit()

    def clear_local_storage(self) -> Browser:
        self.driver.execute_script(
            'window.localStorage.clear();'
        )  # todo: should we catch and ignore errors?
        return self

    def clear_session_storage(self) -> Browser:
        self.driver.execute_script('window.sessionStorage.clear();')
        return self

    # --- Assertable --- #

    # we need this method here in order to make autocompletion work...
    # unfortunately the "base class" version is not enough
    def should(self, condition: BrowserCondition) -> Browser:
        super().should(condition)
        return self

    # --- Deprecated --- #

    def close(self) -> Browser:
        warnings.warn(
            "deprecated; use a `close_current_tab` method instead",
            DeprecationWarning,
        )

        return self.close_current_tab()

    def quit_driver(self) -> None:
        warnings.warn(
            "deprecated; use a `quit` method instead", DeprecationWarning
        )
        self.quit()

    @classmethod
    def wrap(cls, webdriver: WebDriver):
        warnings.warn(
            "deprecated; use a normal constructor style instead: "
            "`Browser(Config(driver=webdriver))`",
            DeprecationWarning,
        )
        return Browser(Config(driver=webdriver))

    def s(self, css_or_xpath_or_by: Union[str, tuple]) -> Element:
        warnings.warn(
            "deprecated; use an `element` method instead: "
            "`browser.element('#foo')`",
            DeprecationWarning,
        )

        return self.element(css_or_xpath_or_by)

    def find(self, css_or_xpath_or_by: Union[str, tuple]) -> Element:
        warnings.warn(
            "deprecated; use an `element` method instead: "
            "`browser.element('#foo')`",
            DeprecationWarning,
        )

        return self.element(css_or_xpath_or_by)

    def ss(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        warnings.warn(
            "deprecated; use an `all` method instead: "
            "`browser.all('.foo')`",
            DeprecationWarning,
        )

        return self.all(css_or_xpath_or_by)

    def find_all(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        warnings.warn(
            "deprecated; use an `all` method instead: "
            "`browser.all('.foo')`",
            DeprecationWarning,
        )

        return self.all(css_or_xpath_or_by)

    def elements(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        warnings.warn(
            "deprecated; use an `all` method instead: "
            "`browser.all('.foo')`",
            DeprecationWarning,
        )

        return self.all(css_or_xpath_or_by)

    def find_elements(self, by=By.ID, value=None):
        warnings.warn(
            "deprecated; use instead: `browser.driver.find_elements(...)`",
            DeprecationWarning,
        )
        return self.driver.find_elements(by, value)

    def find_element(self, by=By.ID, value=None):
        warnings.warn(
            "deprecated; use instead: `browser.driver.find_element(...)`",
            DeprecationWarning,
        )
        return self.driver.find_element(by, value)


# TODO: probably this was needed for migration from 1.0 to 2.0...
from selene.core.conditions import (
    CollectionCondition,
    ElementCondition,
    BrowserCondition,
)


# --- Deprecated --- #  todo: remove in future versions


class SeleneDriver(Browser):
    pass



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re


def is_truthy(something):
    return bool(something) if not something == '' else True


def equals_ignoring_case(expected):
    return lambda actual: str(expected).lower() == str(actual).lower()


def equals(expected, ignore_case=False):
    return (
        lambda actual: expected == actual
        if not ignore_case
        else equals_ignoring_case(expected)
    )


def is_greater_than(expected):
    return lambda actual: actual > expected


def is_greater_than_or_equal(expected):
    return lambda actual: actual >= expected


def is_less_than(expected):
    return lambda actual: actual < expected


def is_less_than_or_equal(expected):
    return lambda actual: actual <= expected


def includes_ignoring_case(expected):
    return lambda actual: str(expected).lower() in str(actual).lower()


def includes(expected, ignore_case=False):
    def fn(actual):
        try:
            return (
                expected in actual
                if not ignore_case
                else includes_ignoring_case(expected)
            )
        except TypeError:
            return False

    return fn


def includes_word_ignoring_case(expected):
    return lambda actual: str(expected).lower() in re.split(
        r'\s+', str(actual).lower()
    )


def includes_word(expected, ignore_case=False):
    return (
        lambda actual: expected in re.split(r'\s+', actual)
        if not ignore_case
        else includes_ignoring_case(expected)
    )


# will not work with empty seqs :( todo: fix
# currently we use it only for non-empty seqs taking this into account
seq_compare_by = (
    lambda f: lambda x=None, *xs: lambda y=None, *ys: True
    if x is None and y is None
    else bool(f(x)(y)) and seq_compare_by(f)(*xs)(*ys)
)


# def seq_compare_by_2(f):
#     def fn(x, *xs):
#         def fn(y, *ys):
#             return True if x is None and y is None else \
#                 bool(f(x)(y)) and seq_compare_by(f)(*xs or (None, ))(*ys or (None, ))
#         return fn
#     return fn


list_compare_by = lambda f: lambda expected: lambda actual: (
    seq_compare_by(f)(*expected)(*actual)
)


# list_compare_by = lambda f: lambda expected: lambda actual: \
#     seq_compare_by(f)(* expected if expected else (None,))(* actual if actual else (None, ))


equals_to_list = list_compare_by(equals)
equals_by_contains_to_list = list_compare_by(includes)



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import Union, Callable


class TimeoutException(AssertionError):
    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        exception_msg = "Message: %s\n" % self.msg
        return exception_msg


# todo: should we extend it from SeleneError and make lazy in same way?
class ConditionNotMatchedError(AssertionError):
    def __init__(self, message='condition not matched'):
        super().__init__(message)


# todo: should we name it *Error or *Exception?
#       (see
#        https://www.python.org/dev/peps/pep-0008/#exception-names
#        https://www.datacamp.com/community/tutorials/exception-handling-python
#       )
#       should we name it simply Error and allow to import as selene.Error ?
#       should we name it SeleneError and still allow to import as selene.Error?
class _SeleneError(AssertionError):
    def __init__(self, message: Union[str, Callable[[], str]]):
        self._render_message: Callable[[], str] = (
            (lambda: message) if isinstance(message, str) else message
        )

    @property
    def args(self):
        return (self._render_message(),)

    def __str__(self):
        return self._render_message()

    def __repr__(self):
        return f"SeleneError: {self._render_message()}"



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import warnings
from functools import lru_cache
from typing import Union, Optional

from selenium.webdriver.remote.webdriver import WebDriver

from selene.common.fp import pipe, identity
from selene.common.helpers import is_absolute_url
from selene.core.entity import Browser, Collection
from selene.core.configuration import Config
from selene.common.none_object import _NoneObject
from selene.core.exceptions import TimeoutException
from selene.core.wait import Wait
from selene.support.shared.config import SharedConfig
from selene.support.webdriver import WebHelper


class SharedBrowser(Browser):
    def __init__(self, config: SharedConfig):
        self._latest_screenshot = _NoneObject(
            'selene.SharedBrowser._latest_screenshot'
        )
        self._latest_page_source = _NoneObject(
            'selene.SharedBrowser._latest_page_source'
        )
        super().__init__(config)

    @property
    def config(self) -> SharedConfig:
        return self._config

    def with_(self, config: Config = None, **config_as_kwargs) -> Browser:
        return SharedBrowser(self.config.with_(config, **config_as_kwargs))

    def open(self, relative_or_absolute_url: str):
        # this -->
        width = self.config.window_width
        height = self.config.window_height

        if width and height:
            self.driver.set_window_size(int(width), int(height))

        is_absolute = is_absolute_url(relative_or_absolute_url)
        base_url = self.config.base_url
        url = (
            relative_or_absolute_url
            if is_absolute
            else base_url + relative_or_absolute_url
        )
        # <-- part is duplicated,
        # --- originally defined in Browser
        # --- todo: consider refactoring

        self.config.get_or_create_driver().get(url)

        return self

    # --- deprecated --- #

    @property
    def driver(self) -> WebDriver:
        webdriver: WebDriver = self.config.driver

        def return_driver(this) -> WebDriver:
            warnings.warn(
                'deprecated; use `browser.driver` over `browser.driver()`',
                DeprecationWarning,
            )
            return webdriver

        webdriver.__class__.__call__ = return_driver

        return webdriver

    def save_screenshot(self, file: str = None):
        # warnings.warn('browser.save_screenshot might be deprecated', FutureWarning)

        if not file:
            file = self.config.generate_filename(suffix='.png')
        if file and not file.lower().endswith('.png'):
            file = os.path.join(file, f'{next(self.config.counter)}.png')
        folder = os.path.dirname(file)
        if not os.path.exists(folder) and folder:
            os.makedirs(folder)
        # todo: refactor to catch errors smartly in get_screenshot_as_file. or not needed?
        self.config.last_screenshot = WebHelper(self.driver).save_screenshot(
            file
        )

        return self.config.last_screenshot

    @property
    def last_screenshot(self) -> str:
        return self.config.last_screenshot

    @property
    def latest_screenshot(self) -> str:
        warnings.warn(
            'deprecated, use browser.last_screenshot property',
            DeprecationWarning,
        )

        class CallableString(str):
            def __new__(cls, value):
                obj = str.__new__(cls, value)
                obj._value = value
                return obj

            def __call__(self, *args, **kwargs):
                warnings.warn(
                    'browser.latest_screenshot() is deprecated, '
                    'use browser.last_screenshot as a property. ',
                    DeprecationWarning,
                )
                return self[:]

            def __bool__(self):
                return bool(self._value)

        return CallableString(self.last_screenshot)

    # todo: consider moving this to browser command.save_page_source(filename)
    def save_page_source(self, file: str = None) -> Optional[str]:
        # warnings.warn('browser.save_page_source(file) might be deprecated in future', FutureWarning)

        if not file:
            file = self.config.generate_filename(suffix='.html')

        saved_file = WebHelper(self.driver).save_page_source(file)

        self.config.last_page_source = saved_file

        return saved_file

    @property
    def last_page_source(self) -> str:
        return self.config.last_page_source

    @property
    def latest_page_source(self):
        warnings.warn(
            'browser.latest_page_source prop is deprecated, use browser.last_page_source',
            DeprecationWarning,
        )
        return self.config.last_page_source

    def quit(self) -> None:
        self.config.reset_driver()

    def quit_driver(self):
        warnings.warn(
            'deprecated; use browser.quit() instead', DeprecationWarning
        )
        self.quit()

    def close(self):
        warnings.warn(
            'deprecated; use browser.close_current_tab() instead',
            DeprecationWarning,
        )
        self.close_current_tab()

    def set_driver(self, webdriver: WebDriver):
        warnings.warn(
            'use config.driver = webdriver '
            'or '
            'config.set_driver = lambda: webdriver '
            'instead',
            DeprecationWarning,
        )

        # noinspection PyDataclass
        self.config.driver = webdriver  # todo: test it

    def open_url(self, absolute_or_relative_url):
        warnings.warn(
            'use browser.open instead of browser.open_url or open_url',
            DeprecationWarning,
        )
        return self.open(absolute_or_relative_url)

    def elements(self, css_or_xpath_or_by: Union[str, tuple]) -> Collection:
        warnings.warn(
            'use browser.all instead of browser.elements or elements',
            DeprecationWarning,
        )
        return self.all(css_or_xpath_or_by)

    def wait_to(self, webdriver_condition, timeout=None, polling=None):
        warnings.warn(
            'use browser.should instead of browser.wait_to or wait_to',
            DeprecationWarning,
        )
        tuned_self = (
            self if timeout is None else self.with_(Config(timeout=timeout))
        )

        return tuned_self.should(webdriver_condition)

    def execute_script(self, script, *args):
        warnings.warn(
            'use browser.driver.execute_script instead of execute_script',
            DeprecationWarning,
        )
        return self.driver.execute_script(script, *args)

    def title(self):
        warnings.warn(
            'use browser.driver.title or browser.get(query.title) instead',
            DeprecationWarning,
        )
        return self.driver.title

    def take_screenshot(self, path=None, filename=None):
        warnings.warn(
            'deprecated; use browser.save_screenshot(filename) instead',
            DeprecationWarning,
        )
        return self.save_screenshot(
            os.path.join(path, filename)
            if path and filename
            else filename or path
        )



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import functools
from typing import TypeVar


T = TypeVar("T")


def identity(it: T) -> T:
    return it


def pipe(*functions):
    """
    pipes functions one by one in the provided order
    i.e. applies arg1, then arg2, then arg3, and so on
    if any arg is None, just skips it
    """
    return (
        functools.reduce(
            lambda f, g: lambda x: f(g(x)) if g else f(x),
            functions[::-1],
            lambda x: x,
        )
        if functions
        else None
    )



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import warnings
from typing import List, Any

from selene.common import predicate
from selene.core import query
from selene.core.condition import Condition
from selene.core.conditions import (
    ElementCondition,
    CollectionCondition,
    BrowserCondition,
)
from selene.core.entity import Collection, Element, Browser

# todo: consider moving to selene.match.element.is_visible, etc...
element_is_visible: Condition[Element] = ElementCondition.raise_if_not(
    'is visible', lambda element: element().is_displayed()
)

element_is_hidden: Condition[Element] = ElementCondition.as_not(
    element_is_visible, 'is hidden'
)

element_is_enabled: Condition[Element] = ElementCondition.raise_if_not(
    'is enabled', lambda element: element().is_enabled()
)

element_is_disabled: Condition[Element] = ElementCondition.as_not(
    element_is_enabled
)

element_is_clickable: Condition[Element] = element_is_visible.and_(
    element_is_enabled
)

element_is_present: Condition[Element] = ElementCondition.raise_if_not(
    'is present in DOM', lambda element: element() is not None
)

element_is_absent: Condition[Element] = ElementCondition.as_not(
    element_is_present
)

element_is_focused: Condition[Element] = ElementCondition.raise_if_not(
    'is focused',
    # todo: change in the following line to element.execute_script or element.config.driver.execute_script
    lambda element: element()
    == element._webdriver.execute_script('return document.activeElement'),
)


def element_has_text(
    expected: str,
    describing_matched_to='has text',
    compared_by_predicate_to=predicate.includes,
) -> Condition[Element]:
    return ElementCondition.raise_if_not_actual(
        describing_matched_to + ' ' + expected,
        query.text,
        compared_by_predicate_to(expected),
    )


def element_has_exact_text(expected: str) -> Condition[Element]:
    return element_has_text(expected, 'has exact text', predicate.equals)


def element_has_js_property(name: str):
    # todo: should we keep simpler but less obvious name - *_has_property ?
    def property_value(element: Element) -> str:
        return element().get_property(name)

    def property_values(collection: Collection) -> List[str]:
        return [element.get_property(name) for element in collection()]

    raw_property_condition = ElementCondition.raise_if_not_actual(
        'has js property ' + name, property_value, predicate.is_truthy
    )

    class ConditionWithValues(ElementCondition):
        def value(self, expected: str) -> Condition[Element]:
            return ElementCondition.raise_if_not_actual(
                f"has js property '{name}' with value '{expected}'",
                property_value,
                predicate.equals(expected),
            )

        def value_containing(self, expected: str) -> Condition[Element]:
            return ElementCondition.raise_if_not_actual(
                f"has js property '{name}' with value containing '{expected}'",
                property_value,
                predicate.includes(expected),
            )

        def values(self, *expected: str) -> Condition[Collection]:
            return CollectionCondition.raise_if_not_actual(
                f"has js property '{name}' with values '{expected}'",
                property_values,
                predicate.equals_to_list(expected),
            )

        def values_containing(self, *expected: str) -> Condition[Collection]:
            return CollectionCondition.raise_if_not_actual(
                f"has js property '{name}' with values containing '{expected}'",
                property_values,
                predicate.equals_by_contains_to_list(expected),
            )

    return ConditionWithValues(
        str(raw_property_condition), raw_property_condition.call
    )


def element_has_css_property(name: str):
    def property_value(element: Element) -> str:
        return element().value_of_css_property(name)

    def property_values(collection: Collection) -> List[str]:
        return [
            element.value_of_css_property(name) for element in collection()
        ]

    raw_property_condition = ElementCondition.raise_if_not_actual(
        'has css property ' + name, property_value, predicate.is_truthy
    )

    class ConditionWithValues(ElementCondition):
        def value(self, expected: str) -> Condition[Element]:
            return ElementCondition.raise_if_not_actual(
                f"has css property '{name}' with value '{expected}'",
                property_value,
                predicate.equals(expected),
            )

        def value_containing(self, expected: str) -> Condition[Element]:
            return ElementCondition.raise_if_not_actual(
                f"has css property '{name}' with value containing '{expected}'",
                property_value,
                predicate.includes(expected),
            )

        def values(self, *expected: str) -> Condition[Collection]:
            return CollectionCondition.raise_if_not_actual(
                f"has css property '{name}' with values '{expected}'",
                property_values,
                predicate.equals_to_list(expected),
            )

        def values_containing(self, *expected: str) -> Condition[Collection]:
            return CollectionCondition.raise_if_not_actual(
                f"has css property '{name}' with values containing '{expected}'",
                property_values,
                predicate.equals_by_contains_to_list(expected),
            )

    return ConditionWithValues(
        str(raw_property_condition), raw_property_condition.call
    )


def element_has_attribute(name: str):
    def attribute_value(element: Element) -> str:
        return element().get_attribute(name)

    def attribute_values(collection: Collection) -> List[str]:
        return [element.get_attribute(name) for element in collection()]

    raw_attribute_condition = ElementCondition.raise_if_not_actual(
        'has attribute ' + name, attribute_value, predicate.is_truthy
    )

    class ConditionWithValues(ElementCondition):
        def value(
            self, expected: str, ignore_case=False
        ) -> Condition[Element]:
            if ignore_case:
                warnings.warn(
                    'ignore_case syntax is experimental and might change in future',
                    FutureWarning,
                )
            return ElementCondition.raise_if_not_actual(
                f"has attribute '{name}' with value '{expected}'",
                attribute_value,
                predicate.equals(expected, ignore_case),
            )

        def value_containing(
            self, expected: str, ignore_case=False
        ) -> Condition[Element]:
            if ignore_case:
                warnings.warn(
                    'ignore_case syntax is experimental and might change in future',
                    FutureWarning,
                )
            return ElementCondition.raise_if_not_actual(
                f"has attribute '{name}' with value containing '{expected}'",
                attribute_value,
                predicate.includes(expected, ignore_case),
            )

        def values(self, *expected: str) -> Condition[Collection]:
            return CollectionCondition.raise_if_not_actual(
                f"has attribute '{name}' with values '{expected}'",
                attribute_values,
                predicate.equals_to_list(expected),
            )

        def values_containing(self, *expected: str) -> Condition[Collection]:
            return CollectionCondition.raise_if_not_actual(
                f"has attribute '{name}' with values containing '{expected}'",
                attribute_values,
                predicate.equals_by_contains_to_list(expected),
            )

    return ConditionWithValues(
        str(raw_attribute_condition), raw_attribute_condition.call
    )


element_is_selected: Condition[Element] = element_has_attribute(
    'elementIsSelected'
)


def element_has_value(expected: str) -> Condition[Element]:
    return element_has_attribute('value').value(expected)


def element_has_value_containing(expected: str) -> Condition[Element]:
    return element_has_attribute('value').value_containing(expected)


def element_has_css_class(expected: str) -> Condition[Element]:
    def class_attribute_value(element: Element) -> str:
        return element().get_attribute('class')

    return ElementCondition.raise_if_not_actual(
        f"has css class '{expected}'",
        class_attribute_value,
        predicate.includes_word(expected),
    )


element_is_blank: Condition[Element] = element_has_exact_text('').and_(
    element_has_value('')
)


def element_has_tag(
    expected: str,
    describing_matched_to='has tag',
    compared_by_predicate_to=predicate.equals,
) -> Condition[Element]:
    return ElementCondition.raise_if_not_actual(
        f'{describing_matched_to} + {expected}',
        query.tag,
        compared_by_predicate_to(expected),
    )


def element_has_tag_containing(expected: str) -> Condition[Element]:
    return element_has_tag(expected, 'has tag containing', predicate.includes)


def _is_collection_empty(collection: Collection) -> bool:
    warnings.warn(
        'match.collection_is_empty or be.empty is deprecated; '
        'use more explicit and obvious have.size(0) instead',
        DeprecationWarning,
    )
    return len(collection()) == 0


collection_is_empty: Condition[Collection] = CollectionCondition.raise_if_not(
    'is empty', _is_collection_empty
)


def collection_has_size(
    expected: int,
    describing_matched_to='has size',
    compared_by_predicate_to=predicate.equals,
) -> Condition[Collection]:
    def size(collection: Collection) -> int:
        return len(collection())

    return CollectionCondition.raise_if_not_actual(
        f'{describing_matched_to} {expected}',
        size,
        compared_by_predicate_to(expected),
    )


def collection_has_size_greater_than(expected: int) -> Condition[Collection]:
    return collection_has_size(
        expected, 'has size greater than', predicate.is_greater_than
    )


def collection_has_size_greater_than_or_equal(
    expected: int,
) -> Condition[Collection]:
    return collection_has_size(
        expected,
        'has size greater than or equal',
        predicate.is_greater_than_or_equal,
    )


def collection_has_size_less_than(expected: int) -> Condition[Collection]:
    return collection_has_size(
        expected, 'has size less than', predicate.is_less_than
    )


def collection_has_size_less_than_or_equal(
    expected: int,
) -> Condition[Collection]:
    return collection_has_size(
        expected,
        'has size less than or equal',
        predicate.is_less_than_or_equal,
    )


# todo: make it configurable whether assert only visible texts or ot
def collection_has_texts(*expected: str) -> Condition[Collection]:
    def visible_texts(collection: Collection) -> List[str]:
        return [
            webelement.text
            for webelement in collection()
            if webelement.is_displayed()
        ]

    return CollectionCondition.raise_if_not_actual(
        f'has texts {expected}',
        visible_texts,
        predicate.equals_by_contains_to_list(expected),
    )


def collection_has_exact_texts(*expected: str) -> Condition[Collection]:
    def visible_texts(collection: Collection) -> List[str]:
        return [
            webelement.text
            for webelement in collection()
            if webelement.is_displayed()
        ]

    return CollectionCondition.raise_if_not_actual(
        f'has exact texts {expected}',
        visible_texts,
        predicate.equals_to_list(expected),
    )


# todo: consider refactoring the code like below by moving outside fns like url, title, etc...
# todo: probably we will do that nevertheless when reusing "commands&queries" inside element class definitions
def browser_has_url(
    expected: str,
    describing_matched_to='has url',
    compared_by_predicate_to=predicate.equals,
) -> Condition[Browser]:
    def url(browser: Browser) -> str:
        return browser.driver.current_url

    return BrowserCondition.raise_if_not_actual(
        f"{describing_matched_to} '{expected}'",
        url,
        compared_by_predicate_to(expected),
    )


def browser_has_url_containing(expected: str) -> Condition[Browser]:
    return browser_has_url(expected, 'has url containing', predicate.includes)


def browser_has_title(
    expected: str,
    describing_matched_to='has title',
    compared_by_predicate_to=predicate.equals,
) -> Condition[Browser]:
    def title(browser: Browser) -> str:
        return browser.driver.title

    return BrowserCondition.raise_if_not_actual(
        f"{describing_matched_to} '{expected}'",
        title,
        compared_by_predicate_to(expected),
    )


def browser_has_title_containing(expected: str) -> Condition[Browser]:
    return browser_has_title(
        expected, 'has title containing', predicate.includes
    )


def browser_has_tabs_number(
    expected: int,
    describing_matched_to='has tabs number',
    compared_by_predicate_to=predicate.equals,
) -> Condition[Browser]:
    def tabs_number(browser: Browser) -> int:
        return len(browser.driver.window_handles)

    return BrowserCondition.raise_if_not_actual(
        f'{describing_matched_to} {expected}',
        tabs_number,
        compared_by_predicate_to(expected),
    )


def browser_has_tabs_number_greater_than(expected: int) -> Condition[Browser]:
    return browser_has_tabs_number(
        expected, 'has tabs number greater than', predicate.is_greater_than
    )


def browser_has_tabs_number_greater_than_or_equal(
    expected: int,
) -> Condition[Browser]:
    return browser_has_tabs_number(
        expected,
        'has tabs number greater than or equal',
        predicate.is_greater_than_or_equal,
    )


def browser_has_tabs_number_less_than(expected: int) -> Condition[Browser]:
    return browser_has_tabs_number(
        expected, 'has tabs number less than', predicate.is_less_than
    )


def browser_has_tabs_number_less_than_or_equal(
    expected: int,
) -> Condition[Browser]:
    return browser_has_tabs_number(
        expected,
        'has tabs number less than or equal',
        predicate.is_less_than_or_equal,
    )


def browser_has_js_returned(
    expected: Any, script: str, *args
) -> Condition[Browser]:
    def script_result(browser: Browser):
        return browser.driver.execute_script(script, *args)

    return BrowserCondition.raise_if_not_actual(
        f'has the ```{script}``` script returned {expected}',
        script_result,
        predicate.equals(expected),
    )



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import warnings


class _ClassPropertyDescriptor:
    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def _classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return _ClassPropertyDescriptor(func)


class BrowserName:
    @_classproperty
    def CHROME(self):
        warnings.warn(
            "might be deprecated, consider using 'chrome' explicitly",
            PendingDeprecationWarning,
        )
        return 'chrome'

    @_classproperty
    def FIREFOX(self):
        warnings.warn(
            "might be deprecated, consider using 'firefox' explicitly",
            PendingDeprecationWarning,
        )
        return 'firefox'

    @_classproperty
    def MARIONETTE(self):
        warnings.warn(
            "might be deprecated, consider using 'firefox' explicitly",
            PendingDeprecationWarning,
        )
        return 'firefox'

    @_classproperty
    def PHANTOMJS(self):
        warnings.warn(
            "might be deprecated, consider using 'phantomjs' explicitly",
            PendingDeprecationWarning,
        )
        return 'chrome'



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import warnings
from typing import Optional

from selenium.webdriver.remote.webdriver import WebDriver

from selene.common.helpers import on_error_return_false


class WebHelper:  # todo: should we make it private? like call it _Help (or think on better name)
    # what about this style for example: ExtendedWebdriver(driver).is_alive() ?
    # over: Help(driver).has_browser_still_alive() ?

    def __init__(self, driver: Optional[WebDriver]):
        self._driver = driver

    def is_browser_still_alive(self):
        return on_error_return_false(lambda: self._driver.title is not None)

    def save_page_source(self, file: str) -> Optional[str]:
        if not file.lower().endswith('.html'):
            warnings.warn(
                "name used for saved pagesource does not match file "
                "type. It should end with an `.html` extension",
                UserWarning,
            )

        html = self._driver.page_source

        try:
            with open(file, 'w', encoding="utf-8") as f:
                f.write(html)
        except OSError:
            return None
        finally:
            del html

        return file

    def save_screenshot(self, file: str) -> Optional[str]:
        if not file.lower().endswith('.png'):
            warnings.warn(
                "name used for saved pagesource does not match file "
                "type. It should end with an `.png` extension",
                UserWarning,
            )

        return file if self._driver.get_screenshot_as_file(file) else None



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Callable, Optional, Union

from selene.common.none_object import _NoneObject
from selene.core.exceptions import TimeoutException
from selenium.webdriver.remote.webdriver import WebDriver

from selene.core.wait import Wait


def _strip_first_underscore(name: str) -> str:
    return name[1:] if name.startswith('_') else name


class Config:
    # todo: add hooks:
    #            hook_wait_command: Callable[[Command], None]
    #            hook_...
    #       consider impementing some hooks as generators...
    #       hook generally should allow to do some processing before and after e.g. command
    #       think on better name depending on hook type (generator vs standard fn)
    def __init__(
        self,
        driver: Optional[Union[WebDriver, Callable[[], WebDriver]]] = None,
        timeout: int = 4,
        hook_wait_failure: Optional[
            Callable[[TimeoutException], Exception]
        ] = None,
        base_url: str = '',
        set_value_by_js: bool = False,
        type_by_js: bool = False,
        wait_for_no_overlap_found_by_js: bool = False,
        window_width: Optional[int] = None,
        window_height: Optional[int] = None,
        log_outer_html_on_failure: bool = False,
    ):

        self._driver = driver
        self._timeout = timeout
        self._hook_wait_failure = hook_wait_failure
        '''
        todo: why we name it as hook_* why not handle_* ?
              what would be proper style?
        '''
        self._base_url = base_url
        self._set_value_by_js = set_value_by_js
        '''
        todo: will it work on mobile? probably no! why then we have it here? o_O
        '''
        self._type_by_js = type_by_js
        self._wait_for_no_overlap_found_by_js = wait_for_no_overlap_found_by_js
        self._window_width = window_width
        self._window_height = window_height
        self._log_outer_html_on_failure = log_outer_html_on_failure
        '''
        todo: should we even keep it as part of core.configuration?
              taking into account that it does not work for mobile
              while we want the core of Selene to be versatile as much as psbl.
              probably we should move it to support.shared.config only
        '''

    def as_dict(self, skip_empty=True):
        return {
            _strip_first_underscore(k): v
            for k, v in self.__dict__.items()
            if not (skip_empty and v is None) and not k.startswith('__')
        }

    def with_(self, config: Config = None, **config_as_kwargs) -> Config:
        return self.__class__(
            **{
                **self.as_dict(),
                **(config.as_dict() if config else {}),
                **config_as_kwargs,
            }
        )

    @property
    def driver(self) -> Union[WebDriver, _NoneObject]:
        return (
            self._driver
            if isinstance(self._driver, WebDriver)
            else (
                self._driver()
                if callable(self._driver)
                else _NoneObject(
                    'expected Callable[[], WebDriver] inside property config.driver'
                )
            )
        )

    @property
    def timeout(self) -> int:
        return self._timeout

    @property
    def hook_wait_failure(self) -> Callable[[TimeoutException], Exception]:
        return self._hook_wait_failure

    # def hook_wait_success(self)
    # def hook_wait(self)

    def wait(self, entity):
        return Wait(
            entity, at_most=self.timeout, or_fail_with=self.hook_wait_failure
        )

    @property
    def base_url(self) -> str:
        return self._base_url

    @property
    def set_value_by_js(self) -> bool:
        return self._set_value_by_js

    @property
    def type_by_js(self) -> bool:
        return self._type_by_js

    @property
    def wait_for_no_overlap_found_by_js(self) -> bool:
        return self._wait_for_no_overlap_found_by_js

    @property
    def window_width(self) -> Optional[int]:
        return self._window_width

    @property
    def window_height(self) -> Optional[int]:
        return self._window_height

    @property
    def log_outer_html_on_failure(self) -> bool:
        return self._log_outer_html_on_failure



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import warnings

from selenium.webdriver.common.by import By


def by(css_selector):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return by_css(css_selector)


def by_css(css_selector):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return (By.CSS_SELECTOR, css_selector)


def by_id(name):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return (By.ID, name)


def by_name(name):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return (By.NAME, name)


def by_link_text(text):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return (By.LINK_TEXT, text)


def by_partial_link_text(text):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return (By.PARTIAL_LINK_TEXT, text)


def by_xpath(xpath):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return (By.XPATH, xpath)


def following_sibling():
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return by_xpath("./following-sibling::*")


def parent():
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return by_xpath("..")


def first_child():
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return by_xpath("./*[1]")


def by_text(element_text):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return by_xpath(
        './/*[text()[normalize-space(.) = '
        + escape_text_quotes_for_xpath(element_text)
        + ']]'
    )


def by_partial_text(element_text):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return by_xpath(
        './/*[text()[contains(normalize-space(.), '
        + escape_text_quotes_for_xpath(element_text)
        + ')]]'
    )


def with_text(element_text):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return by_partial_text(element_text)


def escape_text_quotes_for_xpath(text):
    warnings.warn(
        'deprecated; use by.* from selene.support.by', DeprecationWarning
    )
    return 'concat("", "%s")' % (str("\", '\"', \"".join(text.split('"'))))



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import time
from abc import abstractmethod, ABC
from typing import Generic, Callable, TypeVar, Optional

from selene.core.exceptions import TimeoutException

from selene.common.fp import identity

T = TypeVar('T')
R = TypeVar('R')
E = TypeVar('E')

# todo: not sure, if we need all these Lambda, Proc, Query, Command in python
# todo: they was added just to quickly port selenidejs waiting mechanism
# todo: let's consider removing them... or moving them e.g. to fp

Lambda = Callable[[T], R]
Proc = Callable[[T], None]
Predicate = Callable[[T], bool]


Fn = Callable[[T], R]


# todo: consider moving outside of "wait" module... because there is no direct cohesion with it
class Query(Generic[T, R]):
    def __init__(self, description: str, fn: Callable[[T], R]):
        self._description = description
        self._fn = fn

    def __call__(self, entity: T) -> R:
        return self._fn(entity)

    def __str__(self):
        return self._description


class Command(Query[T, None]):
    pass


# todo: provide sexy fluent implementation via builder, i.e. Wait.the(element).atMost(3).orFailWith(hook)
class Wait(Generic[E]):

    # todo: provide the smallest possible timeout default, something like 1ms
    def __init__(
        self,
        entity: E,
        at_most: int,
        or_fail_with: Optional[Callable[[TimeoutException], Exception]] = None,
    ):
        self._entity = entity
        self._timeout = at_most
        self._hook_failure = or_fail_with or identity

    def at_most(self, timeout: int) -> Wait[E]:
        return Wait(self._entity, timeout, self._hook_failure)

    def or_fail_with(
        self, hook_failure: Optional[Callable[[TimeoutException], Exception]]
    ) -> Wait[E]:

        return Wait(self._entity, self._timeout, hook_failure)

    @property
    def hook_failure(
        self,
    ) -> Optional[Callable[[TimeoutException], Exception]]:
        # todo: hook_failure or failure_hook?
        return self._hook_failure

    # todo: consider renaming to `def to(...)`, though will sound awkward when wait.to(condition)
    def for_(self, fn: Callable[[E], R]) -> R:
        finish_time = time.time() + self._timeout

        while True:
            try:
                return fn(self._entity)
            except Exception as reason:
                if time.time() > finish_time:

                    reason_message = str(reason)

                    reason_string = '{name}: {message}'.format(
                        name=reason.__class__.__name__, message=reason_message
                    )
                    # todo: think on how can we improve logging failures in selene, e.g. reverse msg and stacktrace
                    # stacktrace = getattr(reason, 'stacktrace', None)
                    timeout = self._timeout
                    entity = self._entity

                    failure = TimeoutException(
                        f'''

Timed out after {timeout}s, while waiting for:
{entity}.{fn}

Reason: {reason_string}'''
                    )

                    raise self._hook_failure(failure)

    def until(self, fn: Callable[[E], R]) -> bool:
        try:
            self.for_(fn)
            return True
        except TimeoutException:
            return False

    # todo: do we really need these aliases?
    def command(self, description: str, fn: Callable[[E], None]) -> None:
        self.for_(Command(description, fn))

    def query(self, description: str, fn: Callable[[E], R]) -> R:
        return self.for_(Query(description, fn))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['api',
 'api.base',
 'api.shared',
 'common',
 'conditions',
 'core',
 'selene',
 'selene.api',
 'selene.api.base',
 'selene.api.shared',
 'selene.common',
 'selene.core',
 'selene.support',
 'selene.support.conditions',
 'selene.support.shared',
 'shared',
 'support',
 'support.conditions',
 'support.shared']

package_data = \
{'': ['*']}

install_requires = \
['future', 'selenium==4.1.0', 'webdriver-manager==3.5.4']

setup_kwargs = {
    'name': 'selene',
    'version': '2.0.0b3',
    'description': 'User-oriented browser tests in Python (Selenide port)',
    'long_description': '# Selene - User-oriented Web UI browser tests in Python (Selenide port)\n\n[![tests](https://github.com/yashaka/selene/actions/workflows/tests.yml/badge.svg)](https://github.com/yashaka/selene/actions/workflows/tests.yml)\n[![codecov](https://codecov.io/gh/yashaka/selene/branch/master/graph/badge.svg)](https://codecov.io/gh/yashaka/selene)\n![Free](https://img.shields.io/badge/free-open--source-green.svg)\n[![MIT License](http://img.shields.io/badge/license-MIT-green.svg)](https://github.com/yashaka/selene/blob/master/LICENSE)\n[![Project Template](http://img.shields.io/badge/project-template-9cf.svg)](https://github.com/yashaka/python-web-test)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Downloads](https://pepy.tech/badge/selene)](https://pepy.tech/project/selene)\n\n![GitHub stats in](https://raw.githubusercontent.com/yashaka/selene/traffic/traffic-selene/in_2021.svg)\n![GitHub views](https://raw.githubusercontent.com/yashaka/selene/traffic/traffic-selene/views.svg)\n![GitHub views per week](https://raw.githubusercontent.com/yashaka/selene/traffic/traffic-selene/views_per_week.svg)\n![GitHub clones](https://raw.githubusercontent.com/yashaka/selene/traffic/traffic-selene/clones.svg)\n![GitHub clones per week](https://raw.githubusercontent.com/yashaka/selene/traffic/traffic-selene/clones_per_week.svg)\n\n[![Join telegram chat https://t.me/selene_py](https://img.shields.io/badge/chat-telegram-blue)](https://t.me/selene_py)\n[![Присоединяйся к чату https://t.me/selene_py_ru](https://img.shields.io/badge/%D1%87%D0%B0%D1%82-telegram-red)](https://t.me/selene_py_ru)\n\n[![Sign up for a course http://autotest.how/selene](https://img.shields.io/badge/course-sign_up-blue)](http://autotest.how/selene)\n[![Запишись на курс http://autotest.how/selene-ru](https://img.shields.io/badge/%D0%BD%D0%B0%D0%B1%D0%BE%D1%80-%D0%BD%D0%B0%20%D0%BA%D1%83%D1%80%D1%81-red)](http://autotest.how/selene-ru)\n[![Учи Selene https://leanpub.com/selene-automation-ru](https://img.shields.io/badge/%D0%BA%D0%BD%D0%B8%D0%B3%D0%B0-leanpub-red)](https://leanpub.com/selene-automation-ru)\n[![Реєструйся на курс http://autotest.how/selene-uk](https://img.shields.io/badge/%D0%BD%D0%B0%D0%B1%D1%96%D1%80-%D0%BD%D0%B0_%D0%BA%D1%83%D1%80%D1%81-yellow)](http://autotest.how/selene-uk)\n\n\nMain features:\n\n- **User-oriented API for Selenium Webdriver** (code like speak common English)\n- **Ajax support** (Smart implicit waiting and retry mechanism)\n- **PageObjects support** (all elements are lazy-evaluated objects)\n- **Automatic driver management** (no need to install and setup driver for quick local execution)\n\n\nSelene was inspired by [Selenide](http://selenide.org/) from Java world.\n\nTests with Selene can be built either in a simple straightforward "selenide\' style or with PageObjects composed from Widgets i.e. reusable element components.\n\n## Table of content\n\n* [Versions](#versions)\n    * [Migration Guide](#migration-guide)\n* [Prerequisites](#prerequisites)\n* [Installation](#installation)\n* [Usage](#usage)\n    * [Quick Start](#quick-start)\n    * [Core API](#core-api)\n    * [Automatic Driver and Browser Management](#automatic-driver-and-browser-management)\n    * [Advanced API](#advanced-api)\n* [Tutorials](#tutorials)\n* [Examples](#more-examples)\n* [Contributing](#contributing)\n* [Release Process](#release-process)\n* [Changelog](#changelog)\n\n## Versions\n  \n* Latest recommended version to use is >= [2.0.0b3](https://pypi.org/project/selene/2.0.0b3/)\n  * it\'s a completely new version of selene, with improved API and speed\n  * supports 3.7 <= python <= 3.9,\n  * bundled with Selenium = 4.1\n  * it\'s incompatible with [1.x](https://github.com/yashaka/selene/tree/1.x)\n  * current master branch is pointed to 2.x\n  * yet in alpha/beta stage, refining API, improving "migratability", and testing\n  * it looks pretty stable, most users already upgraded to 2.0 alpha/beta\n\n* Latest version marked as stable is: [1.0.2](https://pypi.org/project/selene/1.0.2/)\n  * it is main version used by most selene users during last 2 years\n  * it was proven to be stable for production use\n  * its sources and corresponding README version can be found at [1.x](https://github.com/yashaka/selene/tree/1.x) branch.\n  * supports python 2.7, 3.5, 3.6, 3.7\n  \nTHIS README DESCRIBES THE USAGE OF THE PRE-RELEASE version of Selene. For older docs look at [1.x](https://github.com/yashaka/selene/tree/1.x) branch.\n  \n### Migration guide\n\nFrom 1.0.1 to 2.0.0aLATEST:\n* upgrade to python 3.7\n* update selene to 2.0.0aLATEST\n  * find&replace the `collection.first()` method from `.first()` to `.first`\n  * ensure all conditions like `text(\'foo\')` are used via `be.*` or `have.*` syntax\n    * example:\n      * find&replace all\n        * `(text(\'foo\'))` to `(have.text(\'foo\'))`\n        * `(visible)` to `(be.visible)`\n      * smarter find&replace (with some manual refactoring)\n        * `.should(x, timeout=y)` to `.with_(timeout=y).should(x)`\n      * and add corresponding imports: `from selene import be, have`\n  * fix another broken imports if available\n  * run tests, read deprecation warnings, and refactor to new style recommended in warning messages\n\n## Prerequisites\n\n[Python >= 3.7](https://www.python.org/downloads/release/python-370/)\n\nGiven [pyenv](https://github.com/pyenv/pyenv) installed, installing needed version of Python is pretty simple:\n\n    $ pyenv install 3.7.3\n    $ pyenv global 3.7.3\n    $ python -V\n    Python 3.7.3\n\n## Installation\n\n### via poetry + pyenv (recommended)\n\nGIVEN [poetry](https://poetry.eustace.io/) and [pyenv](https://github.com/pyenv/pyenv) installed ...\n\nAND\n\n    $ poetry new my-tests-with-selene\n    $ cd my-tests-with-selene\n    $ pyenv local 3.7.3\n\nWHEN latest pre-release recommended version:\n\n    $ poetry add selene --allow-prereleases\n\nWHEN latest stable version:\n\n    $ poetry add selene\n\nTHEN\n\n    $ poetry install\n\n### via pip\n\nLatest recommended pre-release alpha version:\n\n    $ pip install selene --pre\n\nLatest stable version:\n\n    $ pip install selene\n\n### from sources\n\nGIVEN webdriver and webdriver_manager are already installed\n\nTHEN\n\n    $ git clone https://github.com/yashaka/selene.git\n    $ python setup.py install\n\nor using pip:\n\n    $ pip install git+https://github.com/yashaka/selene.git\n    \n\n## Usage\n\n### Quick Start\n\nSimply...\n\n```python\nfrom selene.support.shared import browser\nfrom selene import by, be, have\n\nbrowser.open(\'https://google.com/ncr\')\nbrowser.element(by.name(\'q\')).should(be.blank)\\\n    .type(\'selenium\').press_enter()\nbrowser.all(\'.srg .g\').should(have.size(10))\\\n    .first.should(have.text(\'Selenium automates browsers\'))\n```\n\nOR with custom setup\n\n```python\nfrom selene.support.shared import browser\nfrom selene import by, be, have\n\nbrowser.config.browser_name = \'firefox\'\nbrowser.config.base_url = \'https://google.com\'\nbrowser.config.timeout = 2\n# browser.config.* = ...\n\nbrowser.open(\'/ncr\')\nbrowser.element(by.name(\'q\')).should(be.blank)\\\n    .type(\'selenium\').press_enter()\nbrowser.all(\'.srg .g\').should(have.size(10))\\\n    .first.should(have.text(\'Selenium automates browsers\'))\n```\n\nOR more Selenide from java style:\n\n```python\nfrom selene.support.shared import config, browser\nfrom selene import by, be, have\nfrom selene.support.shared.jquery_style import s, ss\n\n\nconfig.browser_name = \'firefox\'\nconfig.base_url = \'https://google.com\'\nconfig.timeout = 2\n# config.* = ...\n\nbrowser.open(\'/ncr\')\ns(by.name(\'q\')).should(be.blank)\\\n    .type(\'selenium\').press_enter()\nss(\'.srg .g\').should(have.size(10))\\\n    .first.should(have.text(\'Selenium automates browsers\'))\n```\n\n### Core Api\n\nGiven:\n\n```python\nfrom selenium.webdriver import Chrome\n```\n\nAND chromedriver executable available in $PATH\n\nWHEN:\n```python\nfrom selene import Browser, Config\n\nbrowser = Browser(Config(\n    driver=Chrome(),\n    base_url=\'https://google.com\',\n    timeout=2))\n```\n\nAND (uncomment if needed):\n\n```python\n# import atexit\n# atexit.register(browser.quit)\n```\n\nAND:\n\n```python\nbrowser.open(\'/ncr\')\n```\n\nAND:\n\n```python\n# browser.element(\'//*[@name="q"]\')).type(\'selenium\').press_enter()\n# OR...\n\n# browser.element(\'[name=q]\')).type(\'selenium\').press_enter()\n# OR...\n\nfrom selene import by\n\n# browser.element(by.name(\'q\')).type(\'selenium\').press_enter()\n# OR...for total readability\n\nquery = browser.element(by.name(\'q\'))  # actual search doesn\'t start here, the element is "lazy"          \n     # here the actual webelement is found\nquery.type(\'selenium\').press_enter()       \n                      # and here it\'s located again, i.e. the element is "dynamic"\n```\n\nAND (in case we need to filter collection of items by some condition like visibility):\n\n```python\nfrom selene import be\n\nresults = browser.all(\'.srg .g\').filtered_by(be.visible)\n```\n\nTHEN:\n\n```python\nfrom selene import have\n\n# results.should(have.size(10))\n# results.first.should(have.text(\'Selenium automates browsers\'))\n# OR...\n\nresults.should(have.size(10))\\\n    .first.should(have.text(\'Selenium automates browsers\'))\n```\n\nFINALLY (if not registered "atexit" before):\n\n```python\nbrowser.quit()\n```\n\n### Automatic Driver and Browser Management\n\nInstead of:\n\n```python\nfrom selene import Browser, Config\n\nbrowser = Browser(Config(\n    driver=Chrome(),\n    base_url=\'https://google.com\',\n    timeout=2))\n```\nYou can simply use the browser and config instance predefined for you in `selene.support.shared` module:\n\n```python\nfrom selene.support.shared import browser, config\n\n# ... do the same with browser.*\n```\nSo you don\'t need to create you driver instance manually. It will be created for you automatically.\n\nYet, if you need some special case, like working with remote driver, etc., you can still use shared browser object, while providing driver to it through:\n\n```python\nconfig.driver = my_remote_driver_instance\n# or\nbrowser.config.driver = my_remote_driver_instance\n```\n\n### Advanced API\n\nSometimes you might need some extra actions on elements, e.g. for workaround something through js:\n\n```python\nfrom selene import command\n\nbrowser.element(\'#not-in-view\').perform(command.js.scroll_into_view)\n```\n\nProbably you think that will need something like:\n\n```python\nfrom selene import query\n\nproduct_text = browser.element(\'#to-assert-something-non-standard\').get(query.text)\nprice = my_int_from(product_text)\nassert price > 100\n```\n\nBut usually it\'s either better to implement your custom condition:\n\n```python\nbrowser.element(\'#to-assert-something-non-standard\').should(have_in_text_the_int_number_more_than(100))\n```\n\nWhere the `have_in_text_the_int_number_more_than` is your defined custom condition. Such condition-based alternative will be less fragile, because python\'s assert does not have "implicit waiting", like selene\'s should ;)\n\n\nFurthermore, the good test is when you totally control your test data, and instead:\n\n```python\nproduct = browser.element(\'#to-remember-for-future\')\n\nproduct_text_before = product.get(query.text)\nprice_before = my_int_from(product_text_before)\n\n# ... do something\n\nproduct_text_after = product.get(query.text)\nprice_after = my_int_from(product_text_after)\n\nassert price_after > price_before\n```\n\nNormally, better would be to refactor to something like:\n\n```python\nproduct = browser.element(\'#to-remember-for-future\')\n\nproduct.should(have.text(\'100$\'))\n\n# ... do something\n\nproduct.should(have.text(\'125$\'))\n```\nYou might think you need something like:\n\n```python\nfrom selene import query\n\nif browser.element(\'#i-might-say-yes-or-no\').get(query.text) == \'yes\':\n    # do something...\n```\n\nOr:\n\n```python\nfrom selene import query\n\nif browser.all(\'.option\').get(query.size) >= 2:\n    # do something...\n```\n\nMaybe one day, you really find a use case:) But for above cases, probably easier would be:\n\n```python\nif browser.element(\'#i-might-say-yes-or-no\').wait_until(have.text(\'yes\')):\n    # do something\n\n# ...\n\nif browser.all(\'.i-will-appear\').wait_until(have.size_greater_than_or_equal(2)):\n    # do something\n```\n\nOr, by using non-waiting versions, if "you are in a rush:)":\n\n```python\nif browser.element(\'#i-might-say-yes-or-no\').matching(have.text(\'yes\')):\n    # do something\n\n# ...\n\nif browser.all(\'.i-will-appear\').matching(have.size_greater_than_or_equal(2)):\n    # do something\n```\n\n\n## Tutorials\n\nTBD\n\n## More examples\n\n* [Project template](https://github.com/yashaka/python-web-test)\n\nTBD\n\n## Contributing\n\n[see CONTRIBUTING.md](https://github.com/yashaka/selene/blob/master/CONTRIBUTING.md)\n\n## Release Process\n\n[see CONTRIBUTING.md#release-process](https://github.com/yashaka/selene/blob/master/CONTRIBUTING.md#release-process)\n\n## Changelog\n\n[see CHANGELOG.md](https://github.com/yashaka/selene/blob/master/CHANGELOG.md)\n',
    'author': 'Iakiv Kramarenko',
    'author_email': 'yashaka@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://github.com/yashaka/selene/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
